#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'

from gc3libs.backends.transport import LocalTransport

class FakeTransport(LocalTransport):
    """This class allows you to override responses to specific commands.

    The `execute_command` of `FakeTransport` instead of executing a
    command will look into the `self.expected_answer` dictionary and
    find a proper triple `(retcode, output, error)`. If not found, the
    command will be executed by the parent (`LocalTransport`)

    To use this class you need to fill the `self.expected_answer` with
    commands (only the *binary*, not the whole command line) as keys,
    and triples (retcode, output, error) as values. For example:

    >>> t = FakeTransport()
    >>> t.expected_answer['echo'] = (0, 'GC3pie is wonderful', '')
    >>> t.execute_command("echo 'GC3pie is wonderful'")
    (0, 'GC3pie is wonderful', '')
    
    """
    def __init__(self, expected_answer={}):
        LocalTransport.__init__(self)        
        self.expected_answer = expected_answer
    
    def execute_command(self, command):
        """parse the command and return fake output and error codes
        depending on the current suppose status of the job"""
        commands = []
        
        # There isn't an easy way to split a complex pipeline like
        # "cmd1 ; cmd2 && cmd3 || cmd4"
        
        # We do split the command line, however, but it will work only
        # in a simple case. Therefore you have to remember that
        # complex command lines will probably not be handled correctly
        # by this method.
        
        _ = command.split(';')
        for i in _:
            commands.extend(i.strip().split('&&'))
        
        for cmd in commands:
            args = cmd.strip().split()
            if args[0] in self.expected_answer:
                return self.expected_answer[args[0]]
            else:
                continue
        return LocalTransport.execute_command(self, command)


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="faketransport",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   ggeotop.py -- Front-end script for submitting multiple `GEOtop` jobs to SMSCG.
#
#   Copyright (C) 2011, 2012 GC3, University of Zurich
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `GEOtop` jobs to SMSCG.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``ggeotop --help`` for program usage instructions.

Contents of a typical input folder::

    svf.asc
    asp.asc
    slp.asc
    dem.asc
    geotop.inpts
    rad/
    in/
    rec/
    maps/
"""

__version__ = '2.1.0rc2 version (SVN $Revision$)'
# summary of user-visible changes
__changelog__ = """

  2012-01-19:
  * Added mechanism to validate input folder on more detailed pattern
  * compression of input done only on those folder/files known (as
    opposite as before that was done with exclusion)

  2011-11-07:
  * Initial release, forked off the ``gmhc_coev`` sources.
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import ggeotop
    ggeotop.GGeotopScript().run()


# std module imports
import csv
import glob
import math
import os
import posix
import re
import shutil
import sys
import time

import tarfile

from pkg_resources import Requirement, resource_filename

# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask


GEOTOP_INPUT_ARCHIVE = "input.tgz"
GEOTOP_OUTPUT_ARCHIVE = "output.tgz"

## custom application class

class GeotopApplication(Application):
    """
    Custom class to wrap the execution of the ``GEOtop` program.

    For more information about GEOtop, see <http://www.geotop.org/>
    """
    # From GEOtop's "read me" file:
    #
    # RUNNING
    # Run this simulation by calling the executable (GEOtop_1.223_static)
    # and giving the simulation directory as an argument.
    #
    # EXAMPLE
    # ls2:/group/geotop/sim/tmp/000001>./GEOtop_1.223_static ./
    #
    # TERMINATION OF SIMULATION BY GEOTOP
    # When GEOtop terminates due to an internal error, it mostly reports this
    # by writing a corresponding file (_FAILED_RUN or _FAILED_RUN.old) in the
    # simulation directory. When is terminates sucessfully this file is
    # named (_SUCCESSFUL_RUN or _SUCCESSFUL_RUN.old).
    #
    # RESTARTING SIMULATIONS THAT WERE TERMINATED BY THE SERVER
    # When a simulation is started again with the same arguments as described
    # above (RUNNING), then it continues from the last saving point. If
    # GEOtop finds a file indicating a successful/failed run, it terminates.

    application_name = 'geotop'

    def _scan_and_tar(self, simulation_dir):
        try:
            gc3libs.log.debug("Compressing input folder '%s'" % simulation_dir)
            cwd = os.getcwd()
            os.chdir(simulation_dir)
            # check if input archive already present. If so, remove it
            # XXX: improvement: add to the current archive only newer files from 'rec' folder
            if os.path.isfile(GEOTOP_INPUT_ARCHIVE):
                try:
                    os.remove(GEOTOP_INPUT_ARCHIVE)
                except OSError, x:
                    gc3libs.log.error("Failed removing '%s': %s: %s",
                                      GEOTOP_INPUT_ARCHIVE, x.__class__, x.message)
                    pass

            tar = tarfile.open(GEOTOP_INPUT_ARCHIVE, "w:gz", dereference=True)
            tar.add('./geotop.inpts')
            tar.add('./in')
            tar.add('./out')
            # tar.add('./maps')
            # tar.add('./rec')
            # tar.add('./rad')
            # tar.add('./svf.asc')
            # tar.add('./slp.asc')
            # tar.add('./asp.asc')
            # tar.add('./dem.asc')
            tar.close()
            os.chdir(cwd)
            yield (tar.name, GEOTOP_INPUT_ARCHIVE)
        except Exception, x:
            gc3libs.log.error("Failed creating input archive '%s': %s: %s",
                              os.path.join(simulation_dir, GEOTOP_INPUT_ARCHIVE),
                              x.__class__,x.message)
            raise

    def __init__(self, simulation_dir, executable=None, **extra_args):
        # remember for later
        self.simulation_dir = simulation_dir

        inputs = dict(self._scan_and_tar(simulation_dir))

        geotop_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/geotop_wrap.sh")

        inputs[geotop_wrapper_sh] = os.path.basename(geotop_wrapper_sh)

        if executable is not None:
            # use the specified executable
            executable_name = './' + os.path.basename(executable)
            inputs[executable] = os.path.basename(executable)
        else:
            raise NotImplementedError("No RTE for GEOtop defined; please specify an executable!")
            # use the default one provided by the RTE
            executable_name = '/$GEOTOP'

        # set some execution defaults...
        extra_args.setdefault('requested_cores', 1)
        extra_args.setdefault('requested_architecture', Run.Arch.X86_64)
        extra_args.setdefault('requested_walltime', Duration(8, hours))
        # ...and remove excess ones
        extra_args.pop('output_dir', None)
        Application.__init__(
            self,
            # GEOtop requires only one argument: the simulation directory
            # In our case, since all input files are staged to the
            # execution directory, the only argument is fixed to ``.``
            arguments = ['./'+os.path.basename(geotop_wrapper_sh), 'input.tgz', executable_name ],
            inputs = inputs,
            outputs = gc3libs.ANY_OUTPUT,
            # outputs = outputs,
            output_dir = os.path.join(simulation_dir, 'tmp'),
            stdout = 'ggeotop.log',
            join=True,
            tags = [ 'APPS/EARTH/GEOTOP-1.224' ],
            **extra_args)


    def terminated(self):
        """
        Analyze the retrieved output and decide whether to submit
        another run or not, depending on whether tag files named
        ``_SUCCESSFUL_RUN`` or ``_FAILED_RUN`` are found.
        """
        # provisionally set exit code to 99 (resubmit), will override
        # later if the tag files ``_SUCCESSFUL_RUN`` or
        # ``_FAILED_RUN`` are found.
        self.execution.returncode = (0, 99)

        full_tarname = os.path.join(self.output_dir, GEOTOP_OUTPUT_ARCHIVE)

        # check and unpack output archive
        if os.path.isfile(full_tarname):
            # execution somehow terminated.
            # untar archive
            gc3libs.log.info("Expected output archive found in file '%s'",
                             full_tarname)
            try:
                tar = tarfile.open(full_tarname)
                gc3libs.log.debug("Output tarfile '%s' contains: %s",
                                  full_tarname, str.join(', ', tar.getnames()))
                tar.extractall(path=self.simulation_dir)
                tar.close()
                os.remove(full_tarname)
            except Exception, ex:
                gc3libs.log.error("Error opening output archive '%s': %s: %s",
                                  full_tarname, ex.__class__, ex.message)
                pass

        tmp_output_dir = self.output_dir
        exclude = [
            os.path.basename(self.arguments[0]),
            self.stdout,
            self.stderr,
            GEOTOP_OUTPUT_ARCHIVE,
            ]

        if not os.path.isdir(tmp_output_dir):
            # output folder not available
            # log failure and stop here
            gc3libs.log.warning("Output folder '%s' not found. Cannot process any further" % tmp_output_dir)
            # use exit code 100 to indicate total failure
            self.execution.returncode = (0, 100)
        else:
            # move files one level up, except the ones listed in `exclude`
            for entry in os.listdir(tmp_output_dir):
                src_entry = os.path.join(tmp_output_dir, entry)
                gc3libs.log.debug("Considering entry '%s' ...", src_entry)
                # concatenate all output files together
                if entry == self.stdout:
                    gc3libs.utils.cat(src_entry, output=os.path.join(self.simulation_dir, entry), append=True)
                    # try remove it
                    os.remove(src_entry)
                    gc3libs.log.debug("  ... appended to '%s'", os.path.join(self.simulation_dir, entry))
                    continue
                if entry == self.stderr:
                    gc3libs.utils.cat(src_entry, output=os.path.join(self.simulation_dir, entry), append=True)
                    # try remove it
                    os.remove(src_entry)
                    gc3libs.log.debug("  ... appended to '%s'", os.path.join(self.simulation_dir, entry))
                    continue
                if entry in exclude or (entry.startswith('script.') and entry.endswith('.sh')):
                    # delete entry and continue with next one
                    os.remove(src_entry)
                    gc3libs.log.debug("  ... it's a GC3Pie auxiliary file, ignore it!",)
                    continue

                # now really move file one level up
                dest_entry = os.path.join(self.simulation_dir, entry)
                if os.path.exists(dest_entry):
                    # backup with numerical suffix
                    # gc3libs.utils.backup(dest_entry)
                    shutil.rmtree(dest_entry, ignore_errors=True)
                os.rename(os.path.join(tmp_output_dir, entry), dest_entry)
                gc3libs.log.debug("  ... moved to '%s'", os.path.join(dest_entry))
            # os.removedirs(tmp_output_dir)
            shutil.rmtree(tmp_output_dir, ignore_errors=True)

            # search for termination files
            if (os.path.exists(os.path.join(self.simulation_dir, '_SUCCESSFUL_RUN'))
                or os.path.exists(os.path.join(self.simulation_dir, 'out', '_SUCCESSFUL_RUN'))
                # XXX: why are we looking for '.old' files??
                or os.path.exists(os.path.join(self.simulation_dir, '_SUCCESSFUL_RUN.old'))
                or os.path.exists(os.path.join(self.simulation_dir,'out', '_SUCCESSFUL_RUN.old'))):
                self.execution.returncode = (0, posix.EX_OK)
            elif (os.path.exists(os.path.join(self.simulation_dir, '_FAILED_RUN'))
                or os.path.exists(os.path.join(self.simulation_dir, '_FAILED_RUN'))
                # XXX: why are we looking for '.old' files??
                or os.path.exists(os.path.join(self.simulation_dir, 'out', '_FAILED_RUN.old'))
                or os.path.exists(os.path.join(self.simulation_dir, 'out', '_FAILED_RUN.old'))):
                # use exit code 100 to indicate total failure
                self.execution.returncode = (0, 100)
            else:
                # should be resubmitted
                # call _scan_and_tar to create new archive
                # XXX: To consider a better way of handling this
                # at the moment the entire input archive is recreated
                gc3libs.log.info("Updating tar archive for resubmission.")
                inputs = dict(self._scan_and_tar(self.simulation_dir))
                # self._scan_and_tar(self.simulation_dir)


class GeotopTask(RetryableTask, gc3libs.utils.Struct):
    """
    Run ``geotop`` on a given simulation directory until completion.
    """
    def __init__(self, simulation_dir, executable=None, **extra_args):
        RetryableTask.__init__(
            self,
            # actual computational job
            GeotopApplication(simulation_dir, executable, **extra_args),
            # keyword arguments
            **extra_args)

    def retry(self):
        """
        Resubmit a GEOtop application instance iff it exited with code 99.

        *Note:* There is currently no upper limit on the number of
        resubmissions!
        """
        if self.task.execution.exitcode == 99:
            return True
        else:
            return False


## main script class

class GGeotopScript(SessionBasedScript):
    """
Scan the specified INPUT directories recursively for simulation
directories and submit a job for each one found; job progress is
monitored and, when a job is done, its output files are retrieved back
into the simulation directory itself.

A simulation directory is defined as a directory containing a
``geotop.inpts`` file.

The ``ggeotop`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``ggeotop`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GeotopTask,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GeotopTask,
            )

    def setup_options(self):
        self.add_param("-x", "--executable", metavar="PATH", #type=executable_file,
                       dest="executable", default=None,
                       help="Path to the GEOtop executable file.")

        self.add_param("-q", "--summary", metavar="TIME", dest="summary_period",
                       default=1800,
                       help="Creates a summary of the current execution.")

    def parse_args(self):
        """
        Check validity and consistency of command-line options.
        """
        if self.params.executable is None:
            raise gc3libs.exceptions.InvalidUsage(
                "Use the '-x' option to specify a valid path to the GEOtop executable.")
        if not os.path.exists(self.params.executable):
            raise gc3libs.exceptions.InvalidUsage(
                "Path '%s' to the GEOtop executable does not exist;"
                " use the '-x' option to specify a valid one."
                % self.params.executable)
        gc3libs.utils.test_file(self.params.executable, os.R_OK|os.X_OK,
                                gc3libs.exceptions.InvalidUsage)


    def new_tasks(self, extra):
        # input_files = self._search_for_input_files(self.params.args, 'geotop.inpts')

        # the real input to GEOtop are the directories containing `geotop.inpts`
        # as well as 'in' and 'out' fodlers
        for path in self._validate_input_folders(self.params.args):
            # construct GEOtop job
            yield GeotopTask(
                path,                   # path to the directory containing input files
                self.params.executable, # path to the GEOtop executable
                # job name
                jobname=gc3libs.utils.basename_sans(path),
                # extra keyword arguments passed to the constructor,
                # see `GeotopTask.__init__`
                **extra.copy()
                )

    def _validate_input_folders(self, paths):
        """
        Recursively scan each location in list `paths` for files
        matching a glob pattern, and return the set of path names to
        such files.

        By default, the value of `self.input_filename_pattern` is used
        as the glob pattern to match file names against, but this can
        be overridden by specifying an explicit argument `pattern`.
        """
        for path in paths:
            self.log.debug("Now processing input path '%s' ..." % path)
            if os.path.isdir(path):
                # recursively scan for input files
                for dirpath, dirnames, filenames in os.walk(path):
                    if ("geotop.inpts" in filenames
                        and "in" in dirnames
                        and "out" in dirnames
                        and not dirpath.endswith("/tmp")
                        and not dirpath.endswith("~")
                        and not os.path.isfile(os.path.join(dirpath, 'out', '_SUCCESSFUL_RUN'))
                        and not os.path.isfile(os.path.join(dirpath, 'out', '_FAILED_RUN'))
                        ):
                        yield dirpath
                    else:
                        gc3libs.log.warning(
                            "Ignoring path '%s':"
                            " will not be included in the simulation input bundle.",
                            dirpath)
            else:
                gc3libs.log.warning("Ignoring input path '%s': not a directory.",
                                    path)



#!/usr/bin/env python
"""
Top-level interface to Grid functionality.
"""
# Copyright (C) 2009-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.1.0rc2 version (SVN $Revision: 3585 $)'
__date__ = '$Date$'


from fnmatch import fnmatch
import os
import posix
import re
import sys
import time
import ConfigParser
import tempfile
import warnings
warnings.simplefilter("ignore")

from gc3libs.compat._collections import defaultdict

import gc3libs
import gc3libs.debug
from gc3libs import Application, Run, Task
from gc3libs.backends.sge import SgeLrms
from gc3libs.backends.pbs import PbsLrms
from gc3libs.backends.lsf import LsfLrms
from gc3libs.backends.shellcmd import ShellcmdLrms
from gc3libs.authentication import Auth
import gc3libs.exceptions
import gc3libs.utils as utils


class Core:
    """Core operations: submit, update state, retrieve (a
snapshot of) output, cancel job.

Core operations are *blocking*, i.e., they return only after the
operation has successfully completed, or an error has been detected.

Operations are always performed by a `Core` object.
`Core` implements an overlay Grid on the resources
specified in the configuration file.
    """
    def __init__(self, cfg):
        # init auths
        self.auto_enable_auth = cfg.auto_enable_auth

        # init backends
        self._lrms = cfg.make_resources()
        if len(self._lrms) == 0:
            raise gc3libs.exceptions.NoResources(
                "No resources given to initialize `gc3libs.core.Core` object!")


    def get_backend(self, name):
        try:
            return self._lrms[name]
        except KeyError:
            raise gc3libs.exceptions.InvalidResourceName(
                "Cannot find computational resource '%s'" %
                name)


    def select_resource(self, match):
        """
        Alter the configured list of resources, and retain only those
        that satisfy predicate `match`.

        Argument `match` can be:

          - either a function (or a generic callable) that is passed
            each `Resource` object in turn, and should return a
            boolean indicating whether the resources should be kept
            (`True`) or not (`False`);

          - or it can be a string: only resources whose name matches
            (wildcards ``*`` and ``?`` are allowed) are retained.
        """
        for lrms in self._lrms.itervalues():
            try:
                if not match(lrms):
                    lrms.enabled = False
            except:
                if not fnmatch(lrms.name, match):
                    lrms.enabled = False
        return len(self._lrms)


    def free(self, app, **extra_args):
        """
        Free up any remote resources used for the execution of `app`.
        In particular, this should delete any remote directories and
        files.

        It is an error to call this method if `app.execution.state` is
        anything other than `TERMINATED`: an `InvalidOperation` exception
        will be raised in this case.

        :raise: `gc3libs.exceptions.InvalidOperation` if `app.execution.state`
                differs from `Run.State.TERMINATED`.
        """
        assert isinstance(app, Task), \
            "Core.free: passed an `app` argument which is not a `Task` instance."
        if isinstance(app, Application):
            return self.__free_application(app, **extra_args)
        else:
            # must be a `Task` instance
            return self.__free_task(app, **extra_args)

    def __free_application(self, app, **extra_args):
        """Implementation of `free` on `Application` objects."""
        if app.execution.state not in [ Run.State.TERMINATING, Run.State.TERMINATED ]:
            raise gc3libs.exceptions.InvalidOperation(
                "Attempting to free resources of job '%s',"
                " which is in non-terminal state." % app)

        auto_enable_auth = extra_args.get('auto_enable_auth', self.auto_enable_auth)

        if hasattr(app.execution, 'resource_name'):
            lrms =  self.get_backend(app.execution.resource_name)
            lrms.free(app)
        else:
            gc3libs.log.debug(
                "Core.__free_application(): Application `%s` does not have an "
                "`execution.resource_name` attribute. Assuming it has been "
                "aborted before submission." % str(app))

    def __free_task(self, task, **extra_args):
        """Implementation of `free` on generic `Task` objects."""
        return task.free(**extra_args)


    def submit(self, app, resubmit=False, **extra_args):
        """
        Submit a job running an instance of the given `app`.  Upon
        successful submission, call the `submitted` method on the
        `app` object.

        At the beginning of the submission process, the
        `app.execution` state is reset to ``NEW``; if submission is
        successfull, the task will be in ``SUBMITTED`` or ``RUNNING``
        state when this call returns.

        :raise: `gc3libs.exceptions.InputFileError` if an input file
                does not exist or cannot otherwise be read.
        """
        assert isinstance(app, Task), \
            "Core.submit: passed an `app` argument which is not a `Task` instance."
        if isinstance(app, Application):
            return self.__submit_application(app, resubmit, **extra_args)
        else:
            # must be a `Task` instance
            return self.__submit_task(app, resubmit, **extra_args)

    def __submit_application(self, app, resubmit, **extra_args):
        """Implementation of `submit` on `Application` objects."""

        gc3libs.log.debug("Submitting %s ..." % str(app))

        auto_enable_auth = extra_args.get('auto_enable_auth', self.auto_enable_auth)

        job = app.execution
        if resubmit:
            job.state = Run.State.NEW
        elif job.state != Run.State.NEW:
            return

        lrms_list = filter(lambda x: x.enabled, self._lrms.itervalues())
        if len(lrms_list) == 0:
            raise gc3libs.exceptions.NoResources(
                "Could not initialize any computational resource"
                " - please check log and configuration file.")

        gc3libs.log.debug('Performing brokering ...')
        # decide which resource to use
        # (Resource)[] = (Scheduler).PerformBrokering((Resource)[],(Application))
        _selected_lrms_list = app.compatible_resources(lrms_list)
        gc3libs.log.debug('Application scheduler returned %d matching resources',
                           len(_selected_lrms_list))
        if 0 == len(_selected_lrms_list):
            raise gc3libs.exceptions.NoResources(
                "No available resource can accomodate the application requirements")

        if len(_selected_lrms_list) <= 1:
            # shortcut: no brokering to do, just use what we've got
            updated_resources = _selected_lrms_list
        else:
            # update status of selected resources
            updated_resources = []
            for r in _selected_lrms_list:
                try:
                    # in-place update of resource status
                    gc3libs.log.debug(
                        "Trying to update status of resource '%s' ..."
                        % r.name)
                    r.get_resource_status()
                    updated_resources.append(r)
                except Exception, x:
                    # ignore errors in update, assume resource has a problem
                    # and just drop it
                    gc3libs.log.error("Cannot update status of resource '%s', dropping it."
                                      " See log file for details."
                                      % r.name)
                    gc3libs.log.debug("Got error from get_resource_status(): %s: %s",
                                      x.__class__.__name__, str(x), exc_info=True)

        # sort resources according to Application's preferences
        _selected_lrms_list = app.rank_resources(updated_resources)
        if len(_selected_lrms_list) == 0:
            raise gc3libs.exceptions.LRMSSubmitError(
                "No computational resources left after brokering."
                " Aborting submission of job '%s'" % app)

        exs = [ ]
        # Scheduler.do_brokering returns a sorted list of valid lrms
        for lrms in _selected_lrms_list:
            gc3libs.log.debug("Attempting submission to resource '%s'..."
                              % lrms.name)
            try:
                job.timestamp[Run.State.NEW] = time.time()
                job.info = ("Submitting to '%s' at %s"
                        % (lrms.name,
                           time.ctime(job.timestamp[Run.State.NEW])))
                lrms.submit_job(app)
            except gc3libs.exceptions.LRMSSkipSubmissionToNextIteration, ex:
                gc3libs.log.info(
                    "Submission of job %s delayed" % app)
                exs.append(ex)
                break
            except Exception, ex:
                gc3libs.log.info(
                    "Error in submitting job to resource '%s': %s: %s",
                    lrms.name, ex.__class__.__name__, str(ex),
                    exc_info=True)
                exs.append(ex)
                continue
            gc3libs.log.info("Successfully submitted %s to: %s",
                             str(app), lrms.name)
            job.state = Run.State.SUBMITTED
            job.resource_name = lrms.name
            job.info = ("Submitted to '%s' at %s"
                        % (job.resource_name,
                           time.ctime(job.timestamp[Run.State.SUBMITTED])))
            app.changed = True
            app.submitted()
            # job submitted; return to caller
            return
        # if wet get here, all submissions have failed; call the
        # appropriate handler method if defined
        ex = app.submit_error(exs)
        if isinstance(ex, Exception):
            app.execution.info = ("Submission failed: %s" % str(ex))
            raise ex
        else:
            return

    def __submit_task(self, task, resubmit, **extra_args):
        """Implementation of `submit` on generic `Task` objects."""
        extra_args.setdefault('auto_enable_auth', self.auto_enable_auth)
        task.submit(resubmit, **extra_args)


    def update_job_state(self, *apps, **extra_args):
        """
        Update state of all applications passed in as arguments.

        If keyword argument `update_on_error` is `False` (default),
        then application execution state is not changed in case a
        backend error happens; it is changed to `UNKNOWN` otherwise.

        Note that if state of a job changes, the `Run.state` calls the
        appropriate handler method on the application/task object.

        :raise: `gc3libs.exceptions.InvalidArgument` in case one of
                the passed `Application` or `Task` objects is
                invalid. This can stop updating the state of other
                objects in the argument list.

        :raise: `gc3libs.exceptions.ConfigurationError` if the
                configuration of this `Core` object is invalid or
                otherwise inconsistent (e.g., a resource references a
                non-existing auth section).

        """
        self.__update_application((app for app in apps if isinstance(app, Application)), **extra_args)
        self.__update_task((app for app in apps if not isinstance(app, Application)), **extra_args)

    def __update_application(self, apps, **extra_args):
        """Implementation of `update_job_state` on `Application` objects."""
        update_on_error = extra_args.get('update_on_error', False)
        auto_enable_auth = extra_args.get('auto_enable_auth', self.auto_enable_auth)

        for app in apps:
            state = app.execution.state
            old_state = state
            gc3libs.log.debug("About to update state of application: %s (currently: %s)", app, state)
            try:
                if state not in [ Run.State.NEW,
                                  Run.State.TERMINATING,
                                  Run.State.TERMINATED,
                                  ]:
                    lrms = self.get_backend(app.execution.resource_name)
                    try:
                        state = lrms.update_job_state(app)
                    except Exception, ex:
                        gc3libs.log.debug(
                            "Error getting status of application '%s': %s: %s",
                            app, ex.__class__.__name__, str(ex), exc_info=True)
                        state = Run.State.UNKNOWN
                        # run error handler if defined
                        ex = app.update_job_state_error(ex)
                        if isinstance(ex, Exception):
                            raise ex
                    if state != old_state:
                        app.changed = True
                        # set log information accordingly
                        if (app.execution.state == Run.State.TERMINATING
                            and app.execution.returncode is not None
                            and app.execution.returncode != 0):
                            # there was some error, try to explain
                            app.execution.info = ("Execution failed on resource: %s" % app.execution.resource_name)
                            signal = app.execution.signal
                            if signal in Run.Signals:
                                app.execution.info = ("Abnormal termination: %s" % signal)
                            else:
                                if os.WIFSIGNALED(app.execution.returncode):
                                    app.execution.info = ("Remote job terminated by signal %d" % signal)
                                else:
                                    app.execution.info = ("Remote job exited with code %d"
                                                          % app.execution.exitcode)

                    if state != Run.State.UNKNOWN or update_on_error:
                        app.execution.state = state

            except (gc3libs.exceptions.InvalidArgument,
                    gc3libs.exceptions.ConfigurationError,
                    gc3libs.exceptions.UnrecoverableAuthError,
                    gc3libs.exceptions.FatalError):
                # Unrecoverable; no sense in continuing --
                # pass immediately on to client code and let
                # it handle this...
                raise

            except gc3libs.exceptions.UnknownJob:
                # information about the job is lost, mark it as failed
                app.execution.returncode = (Run.Signals.Lost, -1)
                app.execution.state = Run.State.TERMINATED
                app.changed = True
                continue


            except gc3libs.exceptions.InvalidResourceName, irn:
                # could be the corresponding LRMS has been removed because of an unrecoverable error
                # mark application as state UNKNOWN
                gc3libs.log.warning("Failed while retrieving resource %s from core.Detailed Error message: %s" % (app.execution.resource_name, str(irn)))
                continue

            # XXX: Re-enabled the catch-all clause otherwise the loop stops at the first erroneous iteration
            except Exception, ex:
                if 'GC3PIE_NO_CATCH_ERRORS' in os.environ:
                    # propagate generic exceptions for debugging purposes
                    raise
                else:
                    gc3libs.log.warning("Ignored error in Core.update_job_state(): %s", str(ex))
                    # print again with traceback at a higher log level
                    gc3libs.log.debug("Ignored error in Core.update_job_state(): %s: %s",
                                      ex.__class__.__name__, str(ex), exc_info=True)
                    continue

    def __update_task(self, tasks, **extra_args):
        """Implementation of `update_job_state` on generic `Task` objects."""
        for task in tasks:
            assert isinstance(task, Task), \
                   "Core.update_job_state: passed an argument which is not a `Task` instance."
            task.update_state()


    def fetch_output(self, app, download_dir=None, overwrite=False, **extra_args):
        """
        Retrieve output into local directory `app.output_dir`;
        optional argument `download_dir` overrides this.

        The download directory is created if it does not exist.  If it
        already exists, and the optional argument `overwrite` is
        `False` (default), it is renamed with a `.NUMBER` suffix and a
        new empty one is created in its place.  Otherwise, if
        'overwrite` is `True`, files are downloaded over the ones
        already present.

        If the task is in TERMINATING state, the state is changed to
        `TERMINATED`, attribute `output_dir`:attr: is set to the
        absolute path to the directory where files were downloaded,
        and the `terminated` transition method is called on the `app`
        object.

        Task output cannot be retrieved when `app.execution` is in one
        of the states `NEW` or `SUBMITTED`; an
        `OutputNotAvailableError` exception is thrown in these cases.

        :raise: `gc3libs.exceptions.OutputNotAvailableError` if no
                output can be fetched from the remote job (e.g., the
                Application/Task object is in `NEW` or `SUBMITTED`
                state, indicating the remote job has not started
                running).
        """
        assert isinstance(app, Task), \
            "Core.fetch_output: passed an `app` argument which is not a `Task` instance."
        if isinstance(app, Application):
            self.__fetch_output_application(app, download_dir, overwrite, **extra_args)
        else:
            # generic `Task` object
            self.__fetch_output_task(app, download_dir, overwrite, **extra_args)

    def __fetch_output_application(self, app, download_dir, overwrite, **extra_args):
        """Implementation of `fetch_output` on `Application` objects."""
        job = app.execution
        if job.state in [ Run.State.NEW, Run.State.SUBMITTED ]:
            raise gc3libs.exceptions.OutputNotAvailableError(
                "Output not available: '%s' currently in state '%s'"
                % (app, app.execution.state))

        auto_enable_auth = extra_args.get('auto_enable_auth', self.auto_enable_auth)

        # determine download dir
        download_dir = app._get_download_dir(download_dir)

        # Prepare/Clean download dir
        try:
            if overwrite:
                if not os.path.exists(download_dir):
                    os.makedirs(download_dir)
            else:
                utils.mkdir_with_backup(download_dir)
        except Exception, ex:
            gc3libs.log.error("Failed creating download directory '%s': %s: %s",
                              download_dir, ex.__class__.__name__, str(ex))
            raise

        # download job output
        try:
            lrms = self.get_backend(job.resource_name)
            # self.auths.get(lrms.auth)
            lrms.get_results(app, download_dir)
            # clear previous data staging errors
            if job.signal == Run.Signals.DataStagingFailure:
                job.signal = 0
        except gc3libs.exceptions.InvalidResourceName, ex:
            gc3libs.log.warning(
                "Failed retrieving resource %s from core."
                " Detailed Error message: %s" % (app.execution.resource_name, str(ex)))
            ex = app.fetch_output_error(ex)
            if isinstance(ex, Exception):
                job.info = ("No output could be retrieved: %s" % str(ex))
                raise ex
            else:
                return
        except gc3libs.exceptions.RecoverableDataStagingError, rex:
            job.info = ("Temporary failure when retrieving results: %s."
                        " Ignoring error, try again." % str(rex))
            return
        except gc3libs.exceptions.UnrecoverableDataStagingError, ex:
            job.signal = Run.Signals.DataStagingFailure
            ex = app.fetch_output_error(ex)
            if isinstance(ex, Exception):
                job.info = ("No output could be retrieved: %s" % str(ex))
                raise ex
        except Exception, ex:
            ex = app.fetch_output_error(ex)
            if isinstance(ex, Exception):
                raise ex

        # successfully downloaded results
        gc3libs.log.debug("Downloaded output of '%s' (which is in state %s)"
                          % (str(app), job.state))

        app.output_dir = os.path.abspath(download_dir)
        app.changed = True

        if job.state == Run.State.TERMINATING:
            gc3libs.log.debug("Final output of '%s' retrieved" % str(app))
        return Task.fetch_output(app, download_dir)


    def __fetch_output_task(self, task, download_dir, overwrite, **extra_args):
        """Implementation of `fetch_output` on generic `Task` objects."""
        return task.fetch_output(download_dir, overwrite, **extra_args)


    def get_resources(self, **extra_args):
        """
        Return list of resources configured into this `Core` instance.
        """
        return [ lrms for lrms in self._lrms.itervalues() ]


    def kill(self, app, **extra_args):
        """
        Terminate a job.

        Terminating a job in RUNNING, SUBMITTED, or STOPPED state
        entails canceling the job with the remote execution system;
        terminating a job in the NEW or TERMINATED state is a no-op.
        """
        assert isinstance(app, Task), \
            "Core.kill: passed an `app` argument which is not a `Task` instance."
        if isinstance(app, Application):
            self.__kill_application(app, **extra_args)
        else:
            self.__kill_task(app, **extra_args)

    def __kill_application(self, app, **extra_args):
        """Implementation of `kill` on `Application` objects."""
        job = app.execution
        auto_enable_auth = extra_args.get('auto_enable_auth', self.auto_enable_auth)
        try:
            lrms = self.get_backend(job.resource_name)
            lrms.cancel_job(app)
        except AttributeError:
            # A job in state NEW does not have a `resource_name`
            # attribute.
            if job.state != Run.State.NEW:
                raise
        except gc3libs.exceptions.InvalidResourceName, irn:
            gc3libs.log.warning("Failed while retrieving resource %s from core.Detailed Error message: %s" % (app.execution.resource_name, str(irn)))
        gc3libs.log.debug("Setting job '%s' status to TERMINATED"
                          " and returncode to SIGCANCEL" % job)
        app.changed = True
        # setting the state runs the state-transition handlers,
        # which may raise an error -- ignore them, but log nonetheless
        try:
            job.state = Run.State.TERMINATED
        except Exception, ex:
            gc3libs.log.info("Ignoring error in state transition"
                             " since task is being killed: %s",
                             str(ex))
        job.signal = Run.Signals.Cancelled
        job.history.append("Cancelled.")

    def __kill_task(self, task, **extra_args):
        extra_args.setdefault('auto_enable_auth', self.auto_enable_auth)
        task.kill(**extra_args)


    def peek(self, app, what='stdout', offset=0, size=None, **extra_args):
        """
        Download `size` bytes (at `offset` bytes from the start) from
        the remote job standard output or error stream, and write them
        into a local file.  Return file-like object from which the
        downloaded contents can be read.

        If `size` is `None` (default), then snarf all available
        contents of the remote stream from `offset` unto the end.

        The only allowed values for the `what` arguments are the
        strings `'stdout'` and `'stderr'`, indicating that the
        relevant section of the job's standard output resp. standard
        error should be downloaded.
        """
        assert isinstance(app, Task), \
            "Core.peek: passed an `app` argument which is not a `Task` instance."
        if isinstance(app, Application):
            return self.__peek_application(app, what, offset, size, **extra_args)
        else:
            return self.__peek_task(app, what, offset, size, **extra_args)

    def __peek_application(self, app, what, offset, size, **extra_args):
        """Implementation of `peek` on `Application` objects."""
        job = app.execution
        if what == 'stdout':
            remote_filename = job.stdout_filename
        elif what == 'stderr':
            remote_filename = job.stderr_filename
        else:
            raise Error("File name requested to `Core.peek` must be"
                        " 'stdout' or 'stderr', not '%s'" % what)

        # Check if local data available
        if job.state == Run.State.TERMINATED:
            # FIXME: local data could be stale!!
            filename = os.path.join(app.output_dir, remote_filename)
            local_file = open(filename, 'r')
        else:
            # Get authN
            auto_enable_auth = extra_args.get('auto_enable_auth', self.auto_enable_auth)
            lrms = self.get_backend(job.resource_name)
            local_file = tempfile.NamedTemporaryFile(suffix='.tmp', prefix='gc3libs.')
            lrms.peek(app, remote_filename, local_file, offset, size)
            local_file.flush()
            local_file.seek(0)

        return local_file

    def __peek_task(self, task, what, offset, size, **extra_args):
        """Implementation of `peek` on generic `Task` objects."""
        return task.peek(what, offset, size, **extra_args)


    def update_resources(self, **extra_args):
        """
        Update the state of resources configured into this `Core` instance.

        Each resource object in the returned list will have its `updated` attribute
        set to `True` if the update operation succeeded, or `False` if it failed.
        """
        for lrms in self._lrms.itervalues():
            try:
                if not lrms.enabled:
                    continue
                auto_enable_auth = extra_args.get('auto_enable_auth', self.auto_enable_auth)
                resource = lrms.get_resource_status()
                resource.updated = True
            except Exception, ex:
                gc3libs.log.error("Got error while updating resource '%s': %s."
                                  % (lrms.name, str(ex)))
                lrms.updated = False


    def close(self):
        """
        Used to invoke explicitly the destructor on objects
        e.g. LRMS
        """
        for lrms in self._lrms.itervalues():
            lrms.close()


    ## compatibility with the `Engine` interface

    def add(self, task):
        """
        This method is here just to allow `Core` and `Engine` objects
        to be used interchangeably.  It's effectively a no-op, as it makes
        no sense in the synchronous/blocking semantics implemented by `Core`.
        """
        pass


    def remove(self, task):
        """
        This method is here just to allow `Core` and `Engine` objects
        to be used interchangeably.  It's effectively a no-op, as it makes
        no sense in the synchronous/blocking semantics implemented by `Core`.
        """
        pass


# Work around infinite recursion error when trying to compare
# `UserDict` instances which can contain each other.  We know
# that two identical tasks are the same object by
# construction, so let's use this to check.
def _contained(elt, lst):
    i = id(elt)
    for item in lst:
        if i == id(item):
            return True
    return False


class Engine(object):
    """
    Submit tasks in a collection, and update their state until a
    terminal state is reached. Specifically:

      * tasks in `NEW` state are submitted;

      * the state of tasks in `SUBMITTED`, `RUNNING` or `STOPPED` state is updated;

      * when a task reaches `TERMINATED` state, its output is downloaded.

    The behavior of `Engine` instances can be further customized by
    setting the following instance attributes:

      `can_submit`
        Boolean value: if `False`, no task will be submitted.

      `can_retrieve`
        Boolean value: if `False`, no output will ever be retrieved.

      `max_in_flight`
        If >0, limit the number of tasks in `SUBMITTED` or `RUNNING`
        state: if the number of tasks in `SUBMITTED`, `RUNNING` or
        `STOPPED` state is greater than `max_in_flight`, then no new
        submissions will be attempted.

      `max_submitted`
        If >0, limit the number of tasks in `SUBMITTED` state: if the
        number of tasks in `SUBMITTED`, `RUNNING` or `STOPPED` state is
        greater than `max_submitted`, then no new submissions will be
        attempted.

      `output_dir`
        Base directory for job output; if not `None`, each task's
        results will be downloaded in a subdirectory named after the
        task's `permanent_id`.

      `fetch_output_overwrites`
        Default value to pass as the `overwrite` argument to
        :meth:`Core.fetch_output` when retrieving results of a
        terminated task.

    Any of the above can also be set by passing a keyword argument to
    the constructor (assume ``g`` is a `Core`:class: instance)::

      | >>> e = Engine(g, can_submit=False)
      | >>> e.can_submit
      | False
    """


    def __init__(self, controller, tasks=list(), store=None,
                 can_submit=True, can_retrieve=True,
                 max_in_flight=0, max_submitted=0,
                 output_dir=None, fetch_output_overwrites=False):
        """
        Create a new `Engine` instance.  Arguments are as follows:

        :param controller:
          A `gc3libs.Core` instance, that will be used to operate on
          tasks.  This is the only required argument.

        :param list apps:
          Initial list of tasks to be managed by this Engine.  Tasks can
          be later added and removed with the `add` and `remove`
          methods (which see).  Defaults to the empty list.

        :param store:
          An instance of `gc3libs.persistence.Store`, or `None`.  If
          not `None`, it will be used to persist tasks after each
          iteration; by default no store is used so no task state is
          persisted.

        :param can_submit:
        :param can_retrieve:
        :param max_in_flight:
        :param max_submitted:
          Optional keyword arguments; see `Engine` for a description.
        """
        # internal-use attributes
        self._new = []
        self._in_flight = []
        self._stopped = []
        self._terminating = []
        self._terminated = []
        self._to_kill = []
        self._core = controller
        self._store = store
        for task in tasks:
            self.add(task)
        # public attributes
        self.can_submit = can_submit
        self.can_retrieve = can_retrieve
        self.max_in_flight = max_in_flight
        self.max_submitted = max_submitted
        self.output_dir = output_dir
        self.fetch_output_overwrites = fetch_output_overwrites


    def add(self, task):
        """
        Add `task` to the list of tasks managed by this Engine.
        Adding a task that has already been added to this `Engine`
        instance results in a no-op.
        """
        state = task.execution.state
        if Run.State.NEW == state:
            queue = self._new
        elif state in [Run.State.SUBMITTED, Run.State.RUNNING, Run.State.UNKNOWN]:
            queue = self._in_flight
        elif Run.State.STOPPED == state:
            queue = self._stopped
        elif Run.State.TERMINATING == state:
            queue = self._terminating
        elif Run.State.TERMINATED == state:
            queue = self._terminated
        else:
            raise AssertionError("Unhandled state '%s' in gc3libs.core.Engine." % state)
        if not _contained(task, queue):
            queue.append(task)
            task.attach(self)

    def remove(self, task):
        """Remove a `task` from the list of tasks managed by this Engine."""
        state = task.execution.state
        if Run.State.NEW == state:
            self._new.remove(task)
        elif Run.State.SUBMITTED == state or Run.State.RUNNING == state or Run.State.UNKNOWN == state:
            self._in_flight.remove(task)
        elif Run.State.STOPPED == state:
            self._stopped.remove(task)
        elif Run.State.TERMINATING == state:
            self._terminating.remove(task)
        elif Run.State.TERMINATED == state:
            self._terminated.remove(task)
        else:
            raise AssertionError("Unhandled state '%s' in gc3libs.core.Engine." % state)
        task.detach()


    def progress(self):
        """
        Update state of all registered tasks and take appropriate action.
        Specifically:

          * tasks in `NEW` state are submitted;

          * the state of tasks in `SUBMITTED`, `RUNNING`, `STOPPED` or `UNKNOWN` state is updated;

          * when a task reaches `TERMINATING` state, its output is downloaded.

          * tasks in `TERMINATED` status are simply ignored.

        The `max_in_flight` and `max_submitted` limits (if >0) are
        taken into account when attempting submission of tasks.
        """
        # prepare
        currently_submitted = 0
        currently_in_flight = 0
        if self.max_in_flight > 0:
            limit_in_flight = self.max_in_flight
        else:
            limit_in_flight = utils.PlusInfinity()
        if self.max_submitted > 0:
            limit_submitted = self.max_submitted
        else:
            limit_submitted = utils.PlusInfinity()

        # update status of SUBMITTED/RUNNING tasks before launching
        # new ones, otherwise we would be checking the status of
        # some tasks twice...
        #gc3libs.log.debug("Engine.progress: updating status of tasks [%s]"
        #                  % str.join(', ', [str(task) for task in self._in_flight]))
        transitioned = []
        for index, task in enumerate(self._in_flight):
            try:
                self._core.update_job_state(task)
                if self._store and task.changed:
                    self._store.save(task)
                state = task.execution.state
                if state == Run.State.SUBMITTED:
                    # only real applications need to be counted
                    # against the limit; policy tasks are exempt
                    # (this applies to all similar clause below)
                    if isinstance(task, Application):
                        currently_submitted += 1
                        currently_in_flight += 1
                # elif state == Run.State.RUNNING or state == Run.State.UNKNOWN:
                elif state == Run.State.RUNNING:
                    if isinstance(task, Application):
                        currently_in_flight += 1
                elif state == Run.State.STOPPED:
                    transitioned.append(index) # task changed state, mark as to remove
                    self._stopped.append(task)
                elif state == Run.State.TERMINATING:
                    transitioned.append(index) # task changed state, mark as to remove
                    self._terminating.append(task)
                elif state == Run.State.TERMINATED:
                    transitioned.append(index) # task changed state, mark as to remove
                    self._terminated.append(task)
            except gc3libs.exceptions.ConfigurationError:
                # Unrecoverable; no sense in continuing -- pass
                # immediately on to client code and let it handle
                # this...
                raise
            except Exception, x:
                gc3libs.log.error("Ignoring error in updating state of task '%s': %s: %s"
                                  % (task, x.__class__.__name__, str(x)),
                                  exc_info=True)
        # remove tasks that transitioned to other states
        for index in reversed(transitioned):
            del self._in_flight[index]

        # execute kills and update count of submitted/in-flight tasks
        #gc3libs.log.debug("Engine.progress: killing tasks [%s]"
        #                  % str.join(', ', [str(task) for task in self._to_kill]))
        transitioned = []
        for index, task in enumerate(self._to_kill):
            try:
                old_state = task.execution.state
                self._core.kill(task)
                if self._store:
                    self._store.save(task)
                if old_state == Run.State.SUBMITTED:
                    if isinstance(task, Application):
                        currently_submitted -= 1
                        currently_in_flight -= 1
                elif old_state == Run.State.RUNNING:
                    if isinstance(task, Application):
                        currently_in_flight -= 1
                self._terminated.append(task)
                transitioned.append(index)
            except Exception, x:
                if 'GC3PIE_NO_CATCH_ERRORS' in os.environ:
                    # propagate generic exceptions for debugging purposes
                    raise
                else:
                    gc3libs.log.error("Ignored error in killing task '%s': %s: %s",
                                      task, x.__class__.__name__, str(x))
                    # print again with traceback info at a higher log level
                    gc3libs.log.debug("Ignored error in killing task '%s': %s: %s",
                                      task, x.__class__.__name__, str(x), exc_info=True)
        # remove tasks that transitioned to other states
        for index in reversed(transitioned):
            del self._to_kill[index]

        # update state of STOPPED tasks; again need to make before new
        # submissions, because it can alter the count of in-flight
        # tasks.
        #gc3libs.log.debug("Engine.progress: updating status of stopped tasks [%s]"
        #                  % str.join(', ', [str(task) for task in self._stopped]))
        transitioned = []
        for index, task in enumerate(self._stopped):
            try:
                self._core.update_job_state(task)
                if self._store and task.changed:
                    self._store.save(task)
                state = task.execution.state
                if state in [Run.State.SUBMITTED, Run.State.RUNNING]:
                    if isinstance(task, Application):
                        currently_in_flight += 1
                        if task.execution.state == Run.State.SUBMITTED:
                            currently_submitted += 1
                    self._in_flight.append(task)
                    transitioned.append(index) # task changed state, mark as to remove
                elif state == Run.State.TERMINATING:
                    self._terminating.append(task)
                    transitioned.append(index) # task changed state, mark as to remove
                elif state == Run.State.TERMINATED:
                    self._terminated.append(task)
                    transitioned.append(index) # task changed state, mark as to remove
            except Exception, x:
                gc3libs.log.error("Ignoring error in updating state of STOPPED task '%s': %s: %s"
                                  % (task, x.__class__.__name__, str(x)),
                                  exc_info=True)
        # remove tasks that transitioned to other states
        for index in reversed(transitioned):
            del self._stopped[index]

        # now try to submit NEW tasks
        #gc3libs.log.debug("Engine.progress: submitting new tasks [%s]"
        #                  % str.join(', ', [str(task) for task in self._new]))
        transitioned = []
        if self.can_submit:
            index = 0
            while (currently_submitted < limit_submitted
                   and currently_in_flight < limit_in_flight
                   and index < len(self._new)):
                task = self._new[index]
                # try to submit; go to SUBMITTED if successful, FAILED if not
                if currently_submitted < limit_submitted and currently_in_flight < limit_in_flight:
                    try:
                        self._core.submit(task)
                        if self._store:
                            self._store.save(task)
                        self._in_flight.append(task)
                        transitioned.append(index)
                        if isinstance(task, Application):
                            currently_submitted += 1
                            currently_in_flight += 1
                    except Exception, x:
                        if 'GC3PIE_NO_CATCH_ERRORS' in os.environ:
                            # propagate generic exceptions for debugging purposes
                            raise
                        else:
                            gc3libs.log.error(
                                "Ignored error in submitting task '%s': %s: %s",
                                task, x.__class__.__name__, str(x))
                            # print again with traceback at a higher log level
                            gc3libs.log.debug(
                                "Ignored error in submitting task '%s': %s: %s",
                                task, x.__class__.__name__, str(x), exc_info=True)
                            # record the fact in the task's history
                            task.execution.history(
                                "Submission failed: %s: %s"
                                % (x.__class__.__name__, str(x)))
                index += 1
        # remove tasks that transitioned to SUBMITTED state
        for index in reversed(transitioned):
            del self._new[index]

        # finally, retrieve output of finished tasks
        #gc3libs.log.debug("Engine.progress: fetching output of tasks [%s]"
        #                  % str.join(', ', [str(task) for task in self._terminating]))
        if self.can_retrieve:
            transitioned = []
            for index, task in enumerate(self._terminating):
                # try to get output
                try:
                    self._core.fetch_output(task)
                except gc3libs.exceptions.UnrecoverableDataStagingError, ex:
                    gc3libs.log.error("Error in fetching output of task '%s',"
                                      " will mark it as TERMINATED"
                                      " (with error exit code %d): %s: %s",
                                      task, posix.EX_IOERR,
                                      ex.__class__.__name__, str(ex), exc_info=True)
                    task.execution.returncode = (Run.Signals.DataStagingFailure,
                                                 posix.EX_IOERR)
                    task.execution.state = Run.State.TERMINATED
                    task.changed = True
                except Exception, x:
                    if 'GC3PIE_NO_CATCH_ERRORS' in os.environ:
                        # propagate generic exceptions for debugging purposes
                        raise
                    else:
                        gc3libs.log.error(
                            "Ignored error in fetching output of task '%s': %s: %s",
                            task, x.__class__.__name__, str(x))
                        gc3libs.log.debug(
                            "Ignored error in fetching output of task '%s': %s: %s",
                            task, x.__class__.__name__, str(x), exc_info=True)
                if task.execution.state == Run.State.TERMINATED:
                    self._terminated.append(task)
                    self._core.free(task)
                    transitioned.append(index)
                if self._store and task.changed:
                    self._store.save(task)
            # remove tasks for which final output has been retrieved
            for index in reversed(transitioned):
                del self._terminating[index]


    def stats(self, only=None):
        """

        Return a dictionary mapping each state name into the count of
        tasks in that state. In addition, the following keys are defined:

        * `ok`:  count of TERMINATED tasks with return code 0

        * `failed`: count of TERMINATED tasks with nonzero return code

        * `total`: total count of managed tasks, whatever their state

        If the optional argument `only` is not None, tasks whose
        whose class is not contained in `only` are ignored.
        : param tuple only: Restrict counting to tasks of these classes.

        """
        if only:
            gc3libs.log.debug(
                "Engine.stats: Restricting to object of class '%s'", only.__name__)
        result = defaultdict(lambda: 0)
        if only:
            result[Run.State.NEW] = len([task for task in self._new
                                                       if isinstance(task, only)])
        else:
            result[Run.State.NEW] = len(self._new)
        for task in self._in_flight:
            if only and not isinstance(task, only):
                continue
            state = task.execution.state
            result[state] += 1
        for task in self._stopped:
            if only and not isinstance(task, only):
                continue
            state = task.execution.state
            result[state] += 1
        for task in self._to_kill:
            if only and not isinstance(task, only):
                continue
            # XXX: presumes no task in the `_to_kill` list is TERMINATED
            state = task.execution.state
            result[state] += 1
        if only:
            result[Run.State.TERMINATING] += len([task for task in self._terminating
                                                               if isinstance(task, only)])
        else:
            result[Run.State.TERMINATING] += len(self._terminating)
        if only:
            result[Run.State.TERMINATED] += len([task for task in self._terminated
                                                               if isinstance(task, only)])
        else:
            result[Run.State.TERMINATED] += len(self._terminated)

        # for TERMINATED tasks, compute the number of successes/failures
        for task in self._terminated:
            if only and not isinstance(task, only):
                continue
            if task.execution.returncode == 0:
                result['ok'] += 1
            else:
                # gc3libs.log.debug(
                #     "Task '%s' failed: return code %s (signal %s, exitcode %s)"
                #     % (task, task.execution.returncode,
                #        task.execution.signal, task.execution.exitcode))
                result['failed'] += 1
        result['total'] = (result[Run.State.NEW]
                           + result[Run.State.SUBMITTED]
                           + result[Run.State.RUNNING]
                           + result[Run.State.STOPPED]
                           + result[Run.State.TERMINATING]
                           + result[Run.State.TERMINATED]
                           + result[Run.State.UNKNOWN])
        return result


    # implement a Core-like interface, so `Engine` objects can be used
    # as substitutes for `Core`.

    def free(self, task, **extra_args):
        """
        Proxy for `Core.free`, which see.
        """
        self._core.free(task)


    def submit(self, task, resubmit=False, **extra_args):
        """
        Submit `task` at the next invocation of `perform`.

        The `task` state is reset to ``NEW`` and then added to the
        collection of managed tasks.
        """
        if resubmit:
            task.execution.state = Run.State.NEW
        return self.add(task)


    def update_job_state(self, *tasks, **extra_args):
        """
        Return list of *current* states of the given tasks.  States
        will only be updated at the next invocation of `progress`; in
        particular, no state-change handlers are called as a result of
        calling this method.
        """
        pass


    def fetch_output(self, task, output_dir=None, overwrite=False, **extra_args):
        """
        Enqueue task for later output retrieval.

        .. warning:: FIXME

          The `output_dir` and `overwrite` parameters are currently ignored.

        """
        self.add(task)


    def kill(self, task, **extra_args):
        """
        Schedule a task for killing on the next `progress` run.
        """
        self._to_kill.append(task)


    def peek(self, task, what='stdout', offset=0, size=None, **extra_args):
        """
        Proxy for `Core.peek` (which see).
        """
        self._core.peek(task, what, offset, size, **extra_args)


    def close(self):
        """
        Call explicilty finalize methods on relevant objects
        e.g. LRMS
        """
        self._core.close()

    # Wrapper methods around `Core` to access the backends directly
    # from the `Engine`.

    @utils.same_docstring_as(Core.select_resource)
    def select_resource(self, match):
        return self._core.select_resource(match)

    @utils.same_docstring_as(Core.get_resources)
    def get_resources(self):
        return self._core.get_resources()

    @utils.same_docstring_as(Core.get_backend)
    def get_backend(self, name):
        return self._core.get_backend(name)



#! /usr/bin/env python
#
"""
Implementation of task collections.

Tasks can be grouped into collections, which are tasks themselves,
therefore can be controlled (started/stopped/cancelled) like a single
whole.  Collection classes provided in this module implement the basic
patterns of job group execution; they can be combined to form more
complex workflows.  Hook methods are provided so that derived classes
can implement problem-specific job control policies.
"""
# Copyright (C) 2009-2013 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.1.0rc2 version (SVN $Revision$)'

import time
import os

from gc3libs.compat._collections import defaultdict

from gc3libs import log, Run, Task
import gc3libs.exceptions
import gc3libs.utils


class TaskCollection(Task):
    """
    Base class for all task collections. A "task collection" is a
    group of tasks, that can be managed collectively as a single one.

    A task collection implements the same interface as the `Task`
    class, so you can use a `TaskCollection` everywhere a `Task` is
    required.  A task collection has a `state` attribute, which is an
    instance of `gc3libs.Run.State`; each concrete collection class
    decides how to deduce a collective state based on the individual
    task states.
    """

    def __init__(self, tasks=None, **extra_args):
        if tasks is None:
            self.tasks = [ ]
        else:
            self.tasks = tasks
        Task.__init__(self, **extra_args)

    @gc3libs.utils.defproperty
    def changed():
        """
        Evaluates to `True` if this task or any of its subtasks has been modified
        and should be saved to persistent storage.
        """
        def fget(self):
            if self._changed:
                return True
            for task in self.tasks:
                if '_changed' in task:
                    if task._changed:
                        return True
            return False
        def fset(self, value):
            self._changed = value
        return locals()

    # manipulate the "controller" interface used to control the associated task
    def attach(self, controller):
        """
        Use the given Controller interface for operations on the job
        associated with this task.
        """
        raise NotImplementedError("Called abstract method TaskCollection.attach() - this should be overridden in derived classes.")

    def detach(self):
        for task in self.tasks:
            task.detach()
        Task.detach(self)


    def add(self, task):
        """
        Add a task to the collection.
        """
        raise NotImplementedError("Called abstract method TaskCollection.add() - this should be overridden in derived classes.")


    def remove(self, task):
        """
        Remove a task from the collection.
        """
        self.tasks.remove(task)
        task.detach()


    # task execution manipulation -- these methods should be overriden
    # in derived classes, to implement the desired policy.

    def submit(self, resubmit=False, **extra_args):
        raise NotImplementedError("Called abstract method TaskCollection.submit() - this should be overridden in derived classes.")


    def update_state(self, **extra_args):
        """
        Update the running state of all managed tasks.
        """
        for task in self.tasks:
            self._controller.update_job_state(task, **extra_args)


    def kill(self, **extra_args):
        # XXX: provide default implementation that kills all jobs?
        raise NotImplementedError("Called abstract method TaskCollection.kill() - this should be overridden in derived classes.")


    def fetch_output(self, output_dir=None, overwrite=False, **extra_args):
        # if `output_dir` is not None, it is interpreted as the base
        # directory where to download files; each task will get its
        # own subdir based on its `.persistent_id`
        coll_output_dir = self._get_download_dir(output_dir)
        for task in self.tasks:
            if task.execution.state != Run.State.TERMINATING:
                continue
            if 'output_dir' in task:
                task_output_dir = task.output_dir
            else:
                task_output_dir = task.persistent_id
            # XXX: uses a feature from `os.path.join`: if the second
            # path is absolute, the first path is discarded and the
            # second one is returned unchanged
            task_output_dir = os.path.join(coll_output_dir, task_output_dir)
            self._controller.fetch_output(
                task,
                task_output_dir,
                overwrite,
                **extra_args)
        # if any sub-task is not yet TERMINATED, return the base
        # output directory for the collection...
        for task in self.tasks:
            if task.execution.state != Run.State.TERMINATED:
                return coll_output_dir
        # ...otherwise set the state to TERMINATED
        self.execution.state = Run.State.TERMINATED
        self.changed = True
        return coll_output_dir


    def free(self):
        """
        This method just asks the Engine to free the contained tasks.
        """
        if self._attached:
            for task in self.tasks:
                self._controller.free(task)


    def peek(self, what, offset=0, size=None, **extra_args):
        """
        Raise a `gc3libs.exceptions.InvalidOperation` error, as there
        is no meaningful semantics that can be defined for `peek` into
        a generic collection of tasks.
        """
        # is there any sensible semantic here?
        raise gc3libs.exceptions.InvalidOperation("Cannot `peek()` on a task collection.")

    def progress(self):
        raise NotImplementedError("Called abstract method TaskCollection.progress() - this should be overridden in derived classes.")



    def stats(self, only=None):
        """
        Return a dictionary mapping each state name into the count of
        tasks in that state. In addition, the following keys are defined:

        * `ok`:  count of TERMINATED tasks with return code 0

        * `failed`: count of TERMINATED tasks with nonzero return code

        * `total`: count of managed tasks, whatever their state

        If the optional argument `only` is not None, tasks whose
        class is not contained in `only` are ignored.

        :param tuple only: Restrict counting to tasks of these classes.

        """
        result = defaultdict(lambda: 0)
        for task in self.tasks:
            if only and not isinstance(task, only):
                continue
            state = task.execution.state
            result[state] += 1
            if state == Run.State.TERMINATED:
                if task.execution.returncode == 0:
                    result['ok'] += 1
                else:
                    result['failed'] += 1
        if only:
            result['total'] = len([task for task in self.tasks
                                                 if isinstance(task, only)])
        else:
            result['total'] = len(self.tasks)
        return result


    def terminated(self):
        """
        Called when the job state transitions to `TERMINATED`, i.e.,
        the job has finished execution (with whatever exit status, see
        `returncode`) and the final output has been retrieved.

        Default implementation for `TaskCollection` is to set the
        exitcode to the maximum of the exit codes of its tasks.
        """
        self.execution._exitcode = max(
            task.execution._exitcode for task in self.tasks
            )


class SequentialTaskCollection(TaskCollection):
    """
    A `SequentialTaskCollection` runs its tasks one at a time.

    After a task has completed, the `next` method is called with the
    index of the finished task in the `self.tasks` list; the return
    value of the `next` method is then made the collection
    `execution.state`.  If the returned state is `RUNNING`, then the
    subsequent task is started, otherwise no action is performed.

    The default `next` implementation just runs the tasks in the order
    they were given to the constructor, and sets the state to
    `TERMINATED` when all tasks have been run.
    """

    def __init__(self, tasks, **extra_args):
        # XXX: check that `tasks` is a sequence type
        TaskCollection.__init__(self, tasks, **extra_args)
        self._current_task = 0

    def add(self, task):
        task.detach()
        self.tasks.append(task)

    def attach(self, controller):
        """
        Use the given Controller interface for operations on the job
        associated with this task.
        """
        if self._current_task is not None:
            self.tasks[self._current_task].attach(controller)
        Task.attach(self, controller)

    def kill(self, **extra_args):
        """
        Stop execution of this sequence.  Kill currently-running task
        (if any), then set collection state to TERMINATED.
        """
        if self._current_task is not None:
            self.tasks[self._current_task].kill(**extra_args)
        self.execution.state = Run.State.TERMINATED
        self.execution.returncode = (Run.Signals.Cancelled, -1)
        self.changed = True


    def next(self, done):
        """
        Return the state or task to run when step number `done` is completed.

        This method is called when a task is finished; the `done`
        argument contains the index number of the just-finished task
        into the `self.tasks` list.  In other words, the task that
        just completed is available as `self.tasks[done]`.

        The return value from `next` can be either a task state (i.e.,
        an instance of `Run.State`), or a valid index number for
        `self.tasks`. In the first case:

        - if the return value is `Run.State.TERMINATED`,
          then no other jobs will be run;
        - otherwise, the return value is assigned to `execution.state`
          and the next job in the `self.tasks` list is executed.

        If instead the return value is a (nonnegative) number, then
        tasks in the sequence will be re-run starting from that index.

        The default implementation runs tasks in the order they were
        given to the constructor, and sets the state to TERMINATED
        when all tasks have been run.  This method can (and should) be
        overridden in derived classes to implement policies for serial
        job execution.
        """
        if done == len(self.tasks) - 1:
            return Run.State.TERMINATED
        else:
            return Run.State.RUNNING


    def submit(self, resubmit=False, **extra_args):
        """
        Start the current task in the collection.
        """
        if self._current_task is None:
            self._current_task = 0
        task = self.tasks[self._current_task]
        task.attach(self._controller)
        task.submit(resubmit, **extra_args)
        if task.execution.state == Run.State.NEW:
            # submission failed, state unchanged
            self.execution.state = Run.State.NEW
        elif task.execution.state == Run.State.SUBMITTED:
            self.execution.state = Run.State.SUBMITTED
        else:
            self.execution.state = Run.State.RUNNING
        self.changed = True
        return self.execution.state


    def update_state(self, **extra_args):
        """
        Update state of the collection, based on the jobs' statuses.
        """
        if self._current_task is None:
            # it's either NEW or TERMINATED, no update
            assert self.execution.state in [ Run.State.NEW, Run.State.TERMINATED ]
            pass
        else:
            # update state of current task
            task = self.tasks[self._current_task]
            task.update_state(**extra_args)
            gc3libs.log.debug("Task #%d in state %s"
                             % (self._current_task, task.execution.state))
        # set state based on the state of current task
        if self._current_task == 0 and task.execution.state in [ Run.State.NEW, Run.State.SUBMITTED ]:
            self.execution.state = task.execution.state
        elif (task.execution.state == Run.State.TERMINATED):
            nxt = self.next(self._current_task)
            if nxt in Run.State:
                self.execution.state = nxt
                if self.execution.state not in [ Run.State.STOPPED,
                                                 Run.State.TERMINATED ]:
                    self._current_task += 1
                    self.changed = True
                    next_task = self.tasks[self._current_task]
                    next_task.attach(self._controller)
                    self.submit(resubmit=True)
            else:
                # `nxt` must be a valid index into `self.tasks`
                self._current_task = nxt
                self.submit(resubmit=True)
        else:
            self.execution.state = Run.State.RUNNING
        return self.execution.state


class StagedTaskCollection(SequentialTaskCollection):
    """
    Simplified interface for creating a sequence of Tasks.
    This can be used when the number of Tasks to run is
    fixed and known at program writing time.

    A `StagedTaskCollection` subclass should define methods `stage0`,
    `stage1`, ... up to `stageN` (for some arbitrary value of N positive
    integer).  Each of these `stageN` must return a `Task`:class:
    instance; the task returned by the `stage0` method will be executed
    first, followed by the task returned by `stage1`, and so on.
    The sequence stops at the first N such that `stageN` is not defined.

    The exit status of the whole sequence is the exit status of the
    last `Task` instance run.  However, if any of the `stageX` methods
    returns an integer value instead of a `Task` instance, then the
    sequence stops and that number is used as the sequence exit
    code.

    """
    def __init__(self, **extra_args):
        try:
            first_stage = self.stage0()
            if isinstance(first_stage, Task):
                # init parent class with the initial task
                SequentialTaskCollection.__init__(self, [first_stage], **extra_args)
            elif isinstance(first_stage, (int, long, tuple)):
                # init parent class with no tasks, an dimmediately set the exitcode
                SequentialTaskCollection.__init__(self, [], **extra_args)
                self.execution.returncode = first_stage
                self.execution.state = Run.State.TERMINATED
            else:
                raise AssertionError("Invalid return value from method `stage0()` of"
                                     " `StagedTaskCollection` object %r:"
                                     " must return `Task` instance or number" % self)
        except AttributeError, ex:
            raise AssertionError("Invalid `StagedTaskCollection` instance %r: %s"
                                 % (self, str(ex)))


    def next(self, done):
        # get next stage (1); if none exists, log it and exit
        try:
            next_stage_fn = getattr(self, "stage%d" % (done+1))
        except AttributeError:
            gc3libs.log.debug("StagedTaskCollection '%s' has no stage%d,"
                              " ending sequence now.", self, (done+1))
            self.execution.returncode = self.tasks[done].execution.returncode
            return Run.State.TERMINATED
        # get next stage (2); if we get an error here, something is wrong in the code
        try:
            next_stage = next_stage_fn()
        except AttributeError, err:
            raise AssertionError("Invalid `StagedTaskCollection` instance %r: %s"
                                 % (self, str(err)))
        # add next stage to the collection, or end graciously
        if isinstance(next_stage, Task):
            self.add(next_stage)
            return Run.State.RUNNING
        elif isinstance(next_stage, (int, long, tuple)):
            self.execution.returncode = next_stage
            return Run.State.TERMINATED
        else:
            raise AssertionError("Invalid return value from method `stage%d()` of"
                                 " `StagedTaskCollection` object %r:"
                                 " must return `Task` instance or number"
                                 % (done+1, self))



class ParallelTaskCollection(TaskCollection):
    """
    A `ParallelTaskCollection` runs all of its tasks concurrently.

    The collection state is set to `TERMINATED` once all tasks have
    reached the same terminal status.
    """

    def __init__(self, tasks=None, **extra_args):
        TaskCollection.__init__(self, tasks, **extra_args)


    def _state(self):
        """
        Return the state of the collection.

        For a `ParallelTaskCollection`, the state of dependent jobs is
        computed by looping across the states STOPPED, RUNNING,
        SUBMITTED, TERMINATING, UNKNOWN, TERMINATED, NEW in the order
        given: the first state for which there is at least one job in
        that state is returned as the overall collection state.  As an
        exception, if the collection is a mixture of NEW and
        TERMINATED jobs, then the global state is RUNNING (presuming
        we're in the middle of a computation).
        """
        stats = self.stats()
        if (stats[Run.State.NEW] > 0
            and stats[Run.State.TERMINATED] > 0
            and stats[Run.State.NEW] + stats[Run.State.TERMINATED] == len(self.tasks)):
            # we're in the middle of a computation (there's a mixture
            # of unsubmitted and finished tasks), so let's chalk this
            # up to ``RUNNING`` state
            return Run.State.RUNNING
        for state in [ Run.State.STOPPED,
                       Run.State.RUNNING,
                       Run.State.SUBMITTED,
                       Run.State.UNKNOWN,
                       Run.State.TERMINATING,
                       # if we get here, then all jobs are TERMINATED or all NEW
                       Run.State.TERMINATED,
                       Run.State.NEW,
                       ]:
            if stats[state] > 0:
                return state
        return Run.State.UNKNOWN

    def add(self, task):
        """
        Add a task to the collection.
        """
        task.detach()
        self.tasks.append(task)
        if self._attached:
            task.attach(self._controller)

    def attach(self, controller):
        """
        Use the given Controller interface for operations on the job
        associated with this task.
        """
        for task in self.tasks:
            if not task._attached:
                task.attach(controller)
        Task.attach(self, controller)

    def kill(self, **extra_args):
        """
        Terminate all tasks in the collection, and set collection
        state to `TERMINATED`.
        """
        for task in self.tasks:
            task.kill(**extra_args)
        self.execution.state = Run.State.TERMINATED
        self.execution.returncode = (Run.Signals.Cancelled, -1)
        self.changed = True


    def progress(self):
        """
        Try to advance all jobs in the collection to the next state in
        a normal lifecycle.  Return list of task execution states.
        """
        return [ task.progress() for task in self.tasks ]


    def submit(self, resubmit=False, **extra_args):
        """
        Start all tasks in the collection.
        """
        for task in self.tasks:
            task.submit(resubmit, **extra_args)
        self.execution.state = self._state()


    def update_state(self, **extra_args):
        """
        Update state of all tasks in the collection.
        """
        for task in self.tasks:
            #gc3libs.log.debug("Updating state of %s in collection %s ..."
            #                  % (task, self))
            task.update_state(**extra_args)
        self.execution.state = self._state()
        if self.execution.state == Run.State.TERMINATED:
            self.execution.returncode = (0, 0)
            # set exitcode based on returncode of sub-tasks
            for task in self.tasks:
                if task.execution.returncode != 0:
                    self.execution.exitcode = 1
            # FIXME: incorrectly sets `changed` each time it's called!
            self.changed = True


class ChunkedParameterSweep(ParallelTaskCollection):

    def __init__(self, min_value, max_value, step, chunk_size, **extra_args):
        """
        Like `ParallelTaskCollection`, but generate a sequence of jobs
        with a parameter varying from `min_value` to `max_value` in
        steps of `step`.  Only `chunk_size` jobs are generated at a
        time, to distribute the burden of job creation along the whole run.
        """
        self.min_value = min_value
        self.max_value = max_value
        self.step = step
        self.chunk_size = chunk_size
        self._floor = min(min_value + (chunk_size * step), max_value)
        initial = list()
        for param in range(min_value, self._floor, step):
            initial.append(self.new_task(param))
        # start with the initial chunk of jobs
        ParallelTaskCollection.__init__(self, initial, **extra_args)


    def new_task(self, param, **extra_args):
        """
        Return the `Task` corresponding to the parameter value `param`.

        This method *must* be overridden in subclasses to generate tasks.
        """
        raise NotImplementedError("Abstract method `ChunkedParameterSweep.new_task()` called - this should have been defined in a derived class.")


    # this is called at every cycle
    def update_state(self, **extra_args):
        """
        Like `ParallelTaskCollection.update_state()`,
        but also creates new tasks if less than
        `chunk_size` are running.
        """
        # XXX: proposal, reset chuck_size from self._controller.max_in_flight
        # this is the way to pass new 'max-running' value to the class
        # this creates though, a tigh coupling with 'controller' and maybe
        # limits the flexibility of the class.
        # In this way we obsolete 'chunked_size' as part of the __init__ args
        # if self._controller:
        #     gc3libs.log.info("Updating %s chunk_size from %d to %d" %
        #                      (self.__class__, self.chunk_size, self._controller.max_in_flight))
        #     self.chunk_size =  self._controller.
        # XXX: shall we als could jobs in Run.State.STOPPED ?
        num_running = len([task for task in self.tasks if
                           task.execution.state in  [ Run.State.NEW,
                                                      Run.State.SUBMITTED,
                                                      Run.State.RUNNING ]])
                                                      # Run.State.UNKNOWN ]])
        # add more jobs if we're close to the end
        # XXX: why using 2*self.chunk_size as treshold ?
        # is the idea to submit more jobs once we reach at least 50% of completion ?
        # if num_running < 2*self.chunk_size and self._floor < self.max_value:
        if 2*num_running < self.chunk_size and self._floor < self.max_value:
            # generate more tasks
            top = min(self._floor + (self.chunk_size * self.step), self.max_value)
            for param in range(self._floor, top, self.step):
                self.add(self.new_task(param, **extra_args))
            self._floor = top
            self.changed = True
        return ParallelTaskCollection.update_state(self, **extra_args)


class RetryableTask(Task):
    """
    Wrap a `Task` instance and re-submit it until a specified
    termination condition is met.

    By default, the re-submission upon failure happens iff execution
    terminated with nonzero return code; the failed task is retried up
    to `self.max_retries` times (indefinitely if `self.max_retries` is 0).

    Override the `retry` method to implement a different retryal policy.

    *Note:* The resubmission code is implemented in the
    `terminated`:meth:, so be sure to call it if you override in
    derived classes.
    """

    def __init__(self, task, max_retries=0, **extra_args):
        """
        Wrap `task` and resubmit it until `self.retry()` returns `False`.

        :param Task task: A `Task` instance that should be retried.

        :param int max_retries: Maximum number of times `task` should be
            re-submitted; use 0 for 'no limit'.
        """
        self.max_retries = max_retries
        self.retried = 0
        self.task = task
        Task.__init__(self, **extra_args)

    @gc3libs.utils.defproperty
    def changed():
        """
        Evaluates to `True` if this task or any of its subtasks has been modified
        and should be saved to persistent storage.
        """
        def fget(self):
            return self._changed or self.task.changed
        def fset(self, value):
            self._changed = value
        return locals()

    def __getattr__(self, name):
        """Proxy public attributes of the wrapped task."""
        if name.startswith('_'):
            raise AttributeError(
                "'%s' object has no attribute '%s'"
                % (self.__class__.__name__, name))
        return getattr(self.task, name)

    def retry(self):
        """
        Return `True` or `False`, depending on whether the failed task
        should be re-submitted or not.

        The default behavior is to retry a task iff its execution
        terminated with nonzero returncode and the maximum retry limit
        has not been reached.  If `self.max_retries` is 0, then the
        dependent task is retried indefinitely.

        Override this method in subclasses to implement a different
        policy.
        """
        if (self.task.execution.returncode != 0
            and ((self.max_retries > 0
                  and self.retried < self.max_retries)
                 or self.max_retries == 0)):
            return True
        else:
            return False

    def attach(self, controller):
        # here `Task.attach` is the invocation of the superclass'
        # `attach` method (which attaches *this* object to a controller),
        # while `self.task.attach` is the propagation of the `attach`
        # method to the wrapped task. (Same for `detach` below.)
        Task.attach(self, controller)
        self.task.attach(controller)

    def detach(self):
        # see comment in `attach` above
        Task.detach(self)
        self.task.detach()

    def fetch_output(self, *args, **extra_args):
        self.task.fetch_output(*args, **extra_args)

    def free(self, **extra_args):
        self.task.free(**extra_args)

    def kill(self, **extra_args):
        self.task.kill(**extra_args)

    def peek(self, *args, **extra_args):
        return self.task.peek(*args, **extra_args)

    def submit(self, resubmit=False, **extra_args):
        self.task.submit(**extra_args)
        # immediately update state if submission of managed task was successful;
        # otherwise this task may remain in ``NEW`` state which causes an
        # unwanted resubmission if the managing programs ends or is interrupted
        # just after the submission...
        # XXX: this is a case for a generic publish/subscribe mechanism!
        if self.task.execution.state != Run.State.NEW:
            self.execution.state = self._recompute_state()

    def _recompute_state(self):
        """
        Determine and return the state based on the current state and
        the state of the wrapped task.
        """
        own_state = self.execution.state
        task_state = self.task.execution.state
        if own_state == task_state:
            return own_state
        elif own_state == Run.State.NEW:
            if task_state == Run.State.NEW:
                return Run.State.NEW
            elif task_state in [ Run.State.SUBMITTED,
                                 Run.State.RUNNING,
                                 Run.State.STOPPED,
                                 Run.State.UNKNOWN ]:
                return task_state
            else:
                return Run.State.RUNNING
        elif own_state == Run.State.SUBMITTED:
            if task_state in [ Run.State.NEW, Run.State.SUBMITTED ]:
                return Run.State.SUBMITTED
            elif task_state in [ Run.State.RUNNING,
                                 Run.State.TERMINATING,
                                 Run.State.TERMINATED ]:
                return Run.State.RUNNING
            else:
                return task_state
        elif own_state == Run.State.RUNNING:
            if task_state in [ Run.State.STOPPED, Run.State.UNKNOWN ]:
                return task_state
            else:
                # if task is NEW, SUBMITTED, RUNNING, etc. -- keep our state
                return own_state
        elif own_state in [ Run.State.TERMINATING, Run.State.TERMINATED ]:
            assert task_state == Run.State.TERMINATED
            return Run.State.TERMINATED
        elif own_state in [ Run.State.STOPPED, Run.State.UNKNOWN ]:
            if task_state in [ Run.State.NEW,
                               Run.State.SUBMITTED,
                               Run.State.RUNNING,
                               Run.State.TERMINATING,
                               Run.State.TERMINATED ]:
                return Run.State.RUNNING
            else:
                return own_state
        else:
            # should not happen!
            raise AssertionError("Unhandled own state '%s'"
                                 " in RetryableTask._recompute_state()", own_state)

    def update_state(self):
        """
        Update the state of the dependent task, then resubmit it if it's
        TERMINATED and `self.retry()` is `True`.
        """
        own_state_old = self.execution.state
        self.task.update_state()
        own_state_new = self._recompute_state()
        if (self.task.execution.state == Run.State.TERMINATED and own_state_old != Run.State.TERMINATED):
            self.execution.returncode = self.task.execution.returncode
            if self.retry():
                self.retried += 1
                self.task.submit(resubmit=True)
                own_state_new = Run.State.RUNNING
            else:
                own_state_new = Run.State.TERMINATED
            self.changed = True
        if own_state_new != own_state_old:
            self.execution.state = own_state_new
            self.changed = True


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="dag",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
Authentication support with Grid proxy certificates.
"""
# Copyright (C) 2009-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.1.0rc2 version (SVN $Revision: 3585 $)'

import errno
import getpass
import os
import random
import string
import subprocess
import sys
import time


import gc3libs
from gc3libs.authentication import Auth
import gc3libs.exceptions
import gc3libs.utils
from gc3libs import Default


## detect ARC version

# ARC's default packages install arclib into /opt/nordugrid/lib/pythonX.Y/site-packages;
# add this anyway in case users did not set their PYTHONPATH correctly
sys.path.append('/opt/nordugrid/lib/python%d.%d/site-packages'
                % sys.version_info[:2])
# this is where arc0 libraries are installed from release 11.05
sys.path.append('/usr/lib/pymodules/python%d.%d/'
                % sys.version_info[:2])

# XXX: current behaviour: prefer ARC0 libraries over ARC1
# in current ARC client deployment, we install both ARC1 and ARC0
# libraries. Being ARC1 libraries still unreliable, we preffer using
# ARC0.
arc_flavour = None
try:
    import arclib
    arc_flavour = Default.ARC0_LRMS
except ImportError, ex:
    # gc3libs.log.warning("Could not import ARC0 libraries: %s", str(ex))
    try:
        import arc
        arc_flavour = Default.ARC1_LRMS
    except ImportError, ex:
        # gc3libs.log.warning("Could not import ARC1 libraries: %s", str(ex))
        pass


## random password generator
_random_password_letters = string.ascii_letters + string.digits

def random_password(length=24):
  return str.join('', [random.choice(_random_password_letters) for _ in xrange(length)])


class GridAuth(object):

    def __init__(self, **auth):

        # check which arc flavor is used
        if not arc_flavour:
            gc3libs.log.error("Authentication could not be verified. Failed loading any ARC module")
            # XXX: should stop here ?
        else:
            gc3libs.log.info("Authentication verified using module '%s'" % arc_flavour.upper())

        try:
            # test validity
            assert auth['type'] in ['voms-proxy', 'grid-proxy' ], (
                "Configuration error: Unknown type: %s."
                " Valid types are 'voms-proxy' or 'grid-proxy'"
                % auth['type'])
            assert auth['cert_renewal_method'] in ['manual', 'slcs'], (
                "Configuration error: Unknown cert_renewal_method: %s."
                " Valid types are 'voms-proxy' and 'grid-proxy'"
                % auth['cert_renewal_method'])

            # read `remember_password` setting; default to 'False'
            if 'remember_password' in auth:
                auth['remember_password'] = gc3libs.utils.string_to_boolean(auth['remember_password'])
            else:
                auth['remember_password'] = False

            # read `private_cert_copy` setting; default to 'False'
            if 'private_credentials_copy' in auth:
                auth['private_credentials_copy'] = gc3libs.utils.string_to_boolean(auth['private_credentials_copy'])
            else:
                auth['private_credentials_copy'] = False
            if auth['private_credentials_copy']:
                assert auth['cert_renewal_method'] == 'slcs', (
                    "Configuration error: 'private_credentials_copy'"
                    " can only be used with the 'slcs' certificate renewal.")
                if 'private_copy_directory' not in auth:
                    # reset `private_credentials_copy` as it's useless
                    # w/out the copy directory
                    gc3libs.log.warning(
                        "auth/%s: 'private_credentials_copy' is set,"
                        " but no value for 'private_copy_directory' was passed:"
                        " the setting is ineffective"
                        " and no private copy will be kept.", auth['name'])
                    auth['private_credentials_copy'] = False
                elif not os.path.exists(auth['private_copy_directory']):
                    raise gc3libs.exceptions.ConfigurationError(
                        "Incorrect setting '%s'"
                        " for 'private_copy_directory' in auth/%s:"
                        " directory does not exist."
                        % (auth['private_copy_directory'], auth['name']))
                elif not os.path.isdir(auth['private_copy_directory']):
                    raise gc3libs.exceptions.ConfigurationError(
                        "Incorrect setting '%s'"
                        " for 'private_copy_directory' in auth/%s:"
                        " path does not point to a directory."
                        % (auth['private_copy_directory'], auth['name']))

            self.user_cert_valid = False
            self.proxy_valid = False
            self._expiration_time = 0 # initially set expiration time way back in the past
            self._passwd = None
            self._keypass = None
            self.__dict__.update(auth)

        except AssertionError, x:
            raise gc3libs.exceptions.ConfigurationError('Erroneous configuration parameter: %s' % str(x))


    def check(self):
        gc3libs.log.debug('Checking auth: GridAuth')

        # this is of course skipped the first time `check` is called,
        # as `self._expiration_time` is set to 0 in the ctor
        remaining = int(self._expiration_time - time.time())
        if remaining > gc3libs.Default.PROXY_VALIDITY_THRESHOLD:
            gc3libs.log.debug(
                "Grid credentials assumed valid for another %d seconds,"
                " will not actually check.", remaining)
            return True

        # WARNING: the following code might be counter-intuitive, but
        # it's what we need! When `private_credentials_copy` is in
        # effect, we have to force the *certificate* renewal in order
        # to store it in the private directory as a side-effect.  When
        # `remember_password` is in effect, we choose force the
        # *proxy* renewal in order to store the SWITCHaai password as
        # a side-effect. (We could renew the cert to the same purpose,
        # and that would guarantee 10 days of operations, but
        # `slcs-init` is still slower than `voms-proxy-init`...)

        if self.private_credentials_copy and self._keypass is None:
            gc3libs.log.warning(
                "Need to ask password anyway"
                " because of 'private_credentials_copy=yes' in the config file.")
            # force cert renewal to store password
            self.user_cert_valid = False
        else:
            self.user_cert_valid = (0 != self.get_end_time("usercert"))

        self._expiration_time = self.get_end_time("proxy")
        if self.remember_password and self._passwd is None:
            gc3libs.log.warning(
                "Need the proxy password anyway"
                " because of 'remember_password=yes' in the config file.")
            # force proxy renewal to store password
            self.proxy_valid = False
        else:
            self.proxy_valid = (0 != self._expiration_time)

        if self.type == 'voms-proxy' and not self.is_voms():
            gc3libs.log.error("Failed while verifying voms attributes")
            self.proxy_valid = False
            # XXX: reset expiration time to circumvent caching
            self._expiration_time = 0

        return (self.user_cert_valid and self.proxy_valid)


    def enable(self):
        # User certificate
        new_cert = False
        if not self.user_cert_valid:
            # Obtain username? Depends on type + cert_renewal_method combination.
            if self.cert_renewal_method == 'slcs':
                # Check if aai_username is already set. If not, ask interactively
                try:
                    self.aai_username
                except AttributeError:
                    self.aai_username = raw_input('Insert SWITCHaai username: ')

                # Check if idp is already set.  If not, ask interactively
                try:
                    self.idp
                except AttributeError:
                    self.idp = raw_input('Insert SWITCHaai Identity Provider (use the command `slcs-info` to list them): ')

            if self._passwd is not None:
                shib_passwd = self._passwd
            else:
                # ask passwd interactively
                shib_passwd = getpass.getpass('Insert SWITCHaai password for user %s:' % self.aai_username)

            if self.private_credentials_copy:
                key_passwd = random_password()
            else:
                key_passwd = shib_passwd

            new_cert = self.renew_cert(shib_passwd, key_passwd)
            # save passwds for later use
            if new_cert and self.remember_password:
                self._passwd = shib_passwd
            if new_cert and self.private_credentials_copy:
                self._keypass = key_passwd

        # renew proxy if cert has changed or proxy expired
        if new_cert or not self.proxy_valid:
            # have to renew proxy, check that we have all needed info and passwd
            if self.type == 'voms-proxy':
                # Check if vo is already set.  If not, ask interactively
                try:
                    self.vo
                except AttributeError:
                    self.vo = raw_input('Insert VO name: ' )

            if self._keypass is not None:
                # `private_credentials_copy` in effect, use the stored random passwd
                keypass = self._keypass
            elif self._passwd is not None:
                # `remember_password` in effect, use the stored SWITCHaai passwd
                keypass = self._passwd
            else:
                # ask interactively
                if new_cert:
                    keypass = key_passwd
                else:
                    if self.type == 'voms-proxy':
                        message = 'Insert VOMS proxy password: '
                    elif self.type == 'grid-proxy':
                        message = 'Insert grid proxy password: '
                    keypass = getpass.getpass(message)

            new_proxy = self.renew_proxy(keypass)
            # save passwds for later use
            if new_proxy and self.remember_password and self._passwd is None:
                self._passwd = keypass

        # check that all is OK
        if not self.check():
            raise gc3libs.exceptions.RecoverableAuthError(
                "Temporary failure in enabling Grid authentication."
                " Grid/VOMS proxy is %s."
                " User certificate is %s."
                % (gc3libs.utils.ifelse(self.proxy_valid,
                                        "valid", "invalid"),
                   gc3libs.utils.ifelse(self.user_cert_valid,
                                        "valid", "invalid")))
        return True


    def renew_cert(self, shib_passwd, key_passwd):
        if self.cert_renewal_method == 'manual':
            raise gc3libs.exceptions.UnrecoverableAuthError(
                "User certificate expired and renewal set to 'manual', please renew it.")

        try:
            cmd = [
                'slcs-init',
                '--idp',      self.idp,
                '--user',     self.aai_username,
                ]
            if self.private_credentials_copy:
                cmd.extend(['--storedir', self.private_copy_directory])
            p = subprocess.Popen(cmd,
                                 stdin= subprocess.PIPE,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.STDOUT)
            (stdout, stderr) = p.communicate('%s\n%s\n' %
                                             (shib_passwd, key_passwd))

        except OSError, x:
            if x.errno in [errno.ENOENT, errno.EPERM, errno.EACCES]:
                raise gc3libs.exceptions.UnrecoverableAuthError(
                    "Failed running '%s': %s."
                    " Please verify that it is available on your $PATH and that it actually works."
                    % (str.join(' ', cmd), str(x)))
            else:
                # other error; presume it's wrong password or a
                # network glitch.... so retry later.
                raise gc3libs.exceptions.RecoverableAuthError(
                    "Failed running '%s': %s." % (str.join(' ', cmd), str(x)))

        except Exception, ex:
            # Intercept any other Error that subprocess may raise
            gc3libs.log.debug("Unexpected error in GridAuth: %s: %s"
                              % (ex.__class__.__name__, str(ex)))
            raise gc3libs.exceptions.UnrecoverableAuthError(
                "Error renewing SLCS certificate: %s" % str(ex))

        # `slcs-init` exited with error, assume transient condition (wrong password or so).
        if p.returncode != 0:
            raise gc3libs.exceptions.RecoverableAuthError(
                "Error running '%s': %s."
                " Assuming temporary failure, will retry later."
                % (str.join(' ', cmd), stdout))

        gc3libs.log.info('Created new SLCS certificate.')
        if self.private_credentials_copy:
            os.environ['X509_USER_CERT'] = os.path.join(self.private_copy_directory, 'usercert.pem')
            os.environ['X509_USER_KEY'] = os.path.join(self.private_copy_directory, 'userkey.pem')
        return True


    def renew_proxy(self, passwd, _arc_flavour=None):
        global arc_flavour # module-level param
        if _arc_flavour is None:
            _arc_flavour = arc_flavour

        if _arc_flavour == Default.ARC1_LRMS:
            gc3libs.log.debug(
                "Proxy support in ARC1 libraries is not yet functional. Falling back to ARC0 method.")
            return self.renew_proxy(passwd, Default.ARC0_LRMS)

        if self.type == 'voms-proxy':
            assert _arc_flavour == Default.ARC0_LRMS
            # first make sure existing proxy is properly removed
            # run voms-proxy-destroy. This guarantees that the renewal
            # takes the recorded self.password into account
            gc3libs.log.debug("Executing voms-proxy-destroy ...")
            try:
                p = subprocess.Popen('voms-proxy-destroy', stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
                (stdout, stderr) = p.communicate()
            except Exception, x:
                gc3libs.log.error("Got the following error from 'voms-proxy-destroy',"
                                  " but I'm ignoring it:"
                                  " %s: %s" % (x.__class__,x.message))

            # Try renew voms credential; another interactive command
            gc3libs.log.debug("Trying to get a new proxy by 'voms-proxy-init' ...")
            cmd = ['voms-proxy-init', '-valid', '24:00', '-rfc', '-q', '-pwstdin']
            if self.vo is not None:
                cmd.extend(['-voms', self.vo ])

        elif self.type == 'grid-proxy':
            assert _arc_flavour == Default.ARC0_LRMS
            gc3libs.log.debug("Trying to get a new proxy by 'voms-proxy-init' ...")
            cmd = ['grid-proxy-init', '-valid', '24:00', '-q', '-pwstdin']

        if self.private_credentials_copy:
            # XXX: does `grid-proxy-init` support `-out`? do we care?
            cmd.extend(['-out', os.path.join(self.private_copy_directory, 'proxy.pem')])

        try:
            p1 = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            (stdout, stderr) = p1.communicate("%s\n" % passwd)

            if p1.returncode != 0:
                # Cannot use voms-proxy-init return code as validation of the command.
                # just report the warning
                gc3libs.log.warning("Command '%s' exited with code %d: %s"
                                    % (str.join(' ', cmd), p1.returncode, stdout))

            if self.private_credentials_copy:
                os.environ['X509_USER_PROXY'] = os.path.join(self.private_copy_directory, 'proxy.pem')

        except ValueError, x:
            # FIXME: is this more a programming error ?
            raise gc3libs.exceptions.RecoverableAuthError(str(x))

        except OSError, x:
            if x.errno in [errno.ENOENT, errno.EPERM, errno.EACCES]:
                raise gc3libs.exceptions.UnrecoverableAuthError(
                    "Failed running '%s': %s."
                    " Please verify that this command is available in"
                    " your $PATH and that it works."
                    % (str.join(' ', cmd), str(x)))
            else:
                raise gc3libs.exceptions.UnrecoverableAuthError(
                    "Unrecoverable error in enabling Grid authentication: %s" % str(x))

        except Exception, ex:
            # Intercept any other Error that subprocess may raise
            gc3libs.log.debug("Unhandled error in GridAuth: %s: %s"
                              % (ex.__class__.__name__, str(ex)))
            raise gc3libs.exceptions.UnrecoverableAuthError(str(ex))

        return True

    @staticmethod
    def get_end_time(cert_type):
        global arc_flavour # module-level constant

        if arc_flavour == Default.ARC0_LRMS:
            # use ARC libraries
            try:
                if cert_type == "proxy":
                    cert = arclib.Certificate(arclib.PROXY)
                elif cert_type == "usercert":
                    cert = arclib.Certificate(arclib.USERCERT)
                else:
                    raise UnrecoverableAuthError("Unsupported cert type '%s'" % cert_type)
                expires = cert.Expires().GetTime()
            except arclib.CertificateError:
                # ARClib cannot read/access cert, consider it expired
                # to force renewal
                return 0

        elif arc_flavour == Default.ARC1_LRMS:
            # use ARC1 libraries
            try:
                userconfig = arc.UserConfig()

                if cert_type == "proxy":
                    cert = arc.Credential(userconfig.ProxyPath(), "", "", "")
                elif cert_type == "usercert":
                    cert = arc.Credential(userconfig.CertificatePath(), "", "", "")
                else:
                    raise UnrecoverableAuthError("Unsupported cert type '%s'" % cert_type)
                expires = cert.GetEndTime().GetTime()
            except arc.CredentialError:
                # ARClib cannot read/access cert, consider it expired
                # to force renewal
                return 0

        else:
            # XXX: should this be `AssertionError` instead? (it's a programming bug...)
            raise UnrecoverableAuthError("Wrong ARC flavour specified '%s'" % str(arc))

        if expires < time.time():
            gc3libs.log.info("%s is expired." % cert_type)
            return 0
        else:
            gc3libs.log.info("%s valid until %s.", cert_type,
                             time.strftime("%a, %d %b %Y %H:%M:%S (local time)",
                                           time.localtime(expires)))
            return expires

    @staticmethod
    def is_voms():
        global arc_flavour # module-level constant

        if arc_flavour == Default.ARC0_LRMS:
            try:
                cert = arclib.Certificate(arclib.PROXY)
            except Exception, e:
                gc3libs.log.warning("Ignoring error while checking Voms certificate: %s" % str(e))
                return False
            if cert.GetSN().endswith('proxy'):
                # non RFC compliant proxy
                gc3libs.log.error("Proxy not in RFC compliant format")
                return False
            return True

        elif arc_flavour == Default.ARC1_LRMS:
            try:
                userconfig = arc.UserConfig()
                voms_trust_dn = arc.VOMSTrustList()
                voms_trust_dn.AddRegex(".*")
                voms_attributes = arc.VOMSACInfoVector()
                cert = arc.Credential(userconfig.ProxyPath(), "", "", "")

                # check first RFC compliant voms proxy
                if cert.GetDN().endswith('proxy'):
                    # non RFC compliant proxy
                    gc3libs.log.error("Proxy not in RFC compliant format")
                    return False
                
                if not userconfig.CACertificatesDirectory():
                    cadir = gc3libs.Default.CERTIFICATE_AUTHORITIES_DIR
                else:
                    cadir = userconfig.CACertificatesDirectory()
                # XXX: for VOMS_DIR use same approach as arcproxy
                vomsdir = os.getenv("X509_VOMS_DIR")
                if not vomsdir:
                    vomsdir = gc3libs.Default.VOMS_DIR
                arc.parseVOMSAC(cert, cadir, vomsdir, voms_trust_dn, voms_attributes, True, True)
                if voms_attributes.size() > 0:
                    # XXX: for the moment, check only the first entry
                    voms_attribute = voms_attributes[0]
                    if hasattr(voms_attribute,'voname'):
                        # XXX: Assume this is enough ?
                        gc3libs.log.info("Found valid voms proxy of VOname '%s'" % voms_attribute.voname)
                        return True
                else:
                    # No attributes found
                    gc3libs.log.error("No voms attributes found. Is it a grid-proxy ?")
                    return False
            except Exception, ex:
                gc3libs.log.error("Exception in ARC1 libraries: %s %s" % (ex.__class__,ex.message))
                return False

Auth.register('grid-proxy', GridAuth)
Auth.register('voms-proxy', GridAuth)


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="arc",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



# test_configparser_subprocess.py
#
# -*- coding: utf-8 -*-
#
#  Copyright (C) 2010-2012 GC3, University of Zurich
#
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or (at your
#  option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

# stdlib imports
from cStringIO import StringIO
import os
import shutil
import tempfile
import re

# nose
from nose.tools import istest, nottest, raises, assert_equal

try:
    from nose.tools import assert_is_instance
except ImportError:
    # Python 2.6 does not support assert_is_instance()
    from nose.tools import assert_true
    def assert_is_instance(obj, cls):
        assert_true(isinstance(obj, cls))

# GC3Pie imports
from gc3libs import Run, Application
import gc3libs.config
import gc3libs.core
import gc3libs.template
from gc3libs.quantity import GB, hours
from gc3libs.backends.shellcmd import ShellcmdLrms
from gc3libs.quantity import Memory, Duration

def _setup_config_file(confstr):
    (fd, name) = tempfile.mkstemp()
    f = os.fdopen(fd, 'w+')
    f.write(confstr)
    f.close()
    return name


def test_valid_conf():
    """Test parsing a valid configuration file"""
    tmpfile = _setup_config_file("""
[auth/ssh]
type = ssh
username = gc3pie

[resource/test]
type = shellcmd
auth = ssh
transport = local
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
override = False
    """)
    try:
        cfg = gc3libs.config.Configuration(tmpfile)
        resources = cfg.make_resources()
        # resources are enabled by default
        assert 'test' in resources
        assert_is_instance(resources['test'], ShellcmdLrms)
        # test types
        assert_is_instance(resources['test']['name'],         str)
        assert_is_instance(resources['test']['max_cores_per_job'], int)
        assert_is_instance(resources['test']['max_memory_per_core'], Memory)
        assert_is_instance(resources['test']['max_walltime'], Duration)
        assert_is_instance(resources['test']['max_cores'],    int)
        assert_is_instance(resources['test']['architecture'], set)
        # test parsed values
        assert_equal(resources['test']['name'],            'test')
        assert_equal(resources['test']['max_cores_per_job'],    2)
        assert_equal(resources['test']['max_memory_per_core'],  2*GB)
        assert_equal(resources['test']['max_walltime'],         8*hours)
        assert_equal(resources['test']['max_cores'],            2)
        assert_equal(resources['test']['architecture'],
                                           set([Run.Arch.X86_64]))
    finally:
        os.remove(tmpfile)


def test_invalid_confs():
    """Test reading invalid configuration files"""
    invalid_confs = [
        # syntax error: no section header
        "THIS IS NOT A CONFIG FILE",
        # syntax error: malformed key=value
        """
[auth/ssh]
type - ssh
        """,
        ]
    for n, confstr in enumerate(invalid_confs):
        import_invalid_conf.description = (
            test_invalid_confs.__doc__ + (' #%d (load)' % n))
        yield import_invalid_conf, confstr

        read_invalid_conf.description = (
            test_invalid_confs.__doc__ + (' #%d (merge_file)' % n))
        yield read_invalid_conf,   confstr

        parse_invalid_conf.description = (
            test_invalid_confs.__doc__ + (' #%d (_parse)' % n))
        yield parse_invalid_conf,  confstr

@raises(gc3libs.exceptions.NoConfigurationFile)
def import_invalid_conf(confstr, **extra_args):
    """`load` just ignores invalid input."""
    tmpfile = _setup_config_file(confstr)
    cfg = gc3libs.config.Configuration()
    try:
        cfg.load(tmpfile)
    finally:
        os.remove(tmpfile)
    assert_equal(len(cfg.resources), 0)
    assert_equal(len(cfg.auths), 0)

@raises(gc3libs.exceptions.ConfigurationError)
def read_invalid_conf(confstr, **extra_args):
    """`merge_file` raises a `ConfigurationError` exception on invalid input."""
    tmpfile = _setup_config_file(confstr)
    cfg = gc3libs.config.Configuration()
    try:
        cfg.merge_file(tmpfile)
    finally:
        os.remove(tmpfile)

@raises(gc3libs.exceptions.ConfigurationError)
def parse_invalid_conf(confstr, **extra_args):
    """`_parse` raises a `ConfigurationError` exception on invalid input."""
    cfg = gc3libs.config.Configuration()
    defaults, resources, auths = cfg._parse(StringIO(confstr))


def test_key_renames():
    """Test that `ncores` is renamed to `max_cores` during parse"""
    tmpfile = _setup_config_file("""
[auth/ssh]
type = ssh
username = gc3pie

[resource/test]
type = shellcmd
auth = ssh
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
# `ncores` renamed to `max_cores`
ncores = 77
architecture = x86_64
override = False
    """)
    try:
        cfg = gc3libs.config.Configuration(tmpfile)
        resources = cfg.make_resources()
        assert 'ncores' not in resources['test']
        assert 'max_cores' in resources['test']
        assert_equal(resources['test']['max_cores'], 77)
    finally:
        os.remove(tmpfile)


class TestReadMultiple(object):
    def setUp(self):
        self.f1 = _setup_config_file("""
[resource/localhost]
type = shellcmd
transport = local
architecture = x86_64
seq = 1
        """)
        self.f2 = _setup_config_file("""
[resource/localhost]
type = shellcmd
transport = local
architecture = x86_64
seq = 1
foo = 2
        """)
        self.f3 = _setup_config_file("""
[resource/localhost]
type = shellcmd
transport = local
architecture = x86_64
seq = 3
        """)

    def tearDown(self):
        os.remove(self.f1)
        os.remove(self.f2)
        os.remove(self.f3)

    def test_read_config_multiple(self):
        """Test that `Configuration` aggregates multiple files"""
        cfg = gc3libs.config.Configuration(self.f1, self.f2, self.f3)
        assert_equal(cfg.resources['localhost']['seq'], '3')
        assert_equal(cfg.resources['localhost']['foo'], '2')

    def test_load_multiple(self):
        """Test that `Configuration.load` aggregates multiple files"""
        cfg = gc3libs.config.Configuration()
        cfg.load(self.f1, self.f2, self.f3)
        assert_equal(cfg.resources['localhost']['seq'], '3')
        assert_equal(cfg.resources['localhost']['foo'], '2')

    def test_merge_file_multiple(self):
        """Test that `Configuration.load` aggregates multiple files"""
        cfg = gc3libs.config.Configuration()
        cfg.merge_file(self.f1)
        cfg.merge_file(self.f2)
        cfg.merge_file(self.f3)
        assert_equal(cfg.resources['localhost']['seq'], '3')
        assert_equal(cfg.resources['localhost']['foo'], '2')


@raises(gc3libs.exceptions.NoConfigurationFile)
def test_no_valid_config1():
    """Test that `Configuration.load` raises an exception if called with no arguments"""
    cfg = gc3libs.config.Configuration()
    cfg.load()


@raises(gc3libs.exceptions.NoConfigurationFile)
def test_no_valid_config2():
    """Test that `Configuration.load` raises an exception if none of the arguments is a valid config file"""
    f1 = _setup_config_file("INVALID INPUT")
    f2 = _setup_config_file("INVALID INPUT")
    cfg = gc3libs.config.Configuration()
    try:
        cfg.load(f1, f2)
    finally:
        os.remove(f1)
        os.remove(f2)


def test_valid_architectures():
    """Test that valid architecture strings are parsed correctly"""
    test_cases = [
        # a sample of the architecture strings that we accept
        ('x86_64',     [gc3libs.Run.Arch.X86_64]),
        ('x86 64-bit', [gc3libs.Run.Arch.X86_64]),
        ('64bits x86', [gc3libs.Run.Arch.X86_64]),
        ('amd64',      [gc3libs.Run.Arch.X86_64]),
        ('AMD64',      [gc3libs.Run.Arch.X86_64]),
        ('intel 64',   [gc3libs.Run.Arch.X86_64]),
        ('emt64',      [gc3libs.Run.Arch.X86_64]),
        ('64 bits',    [gc3libs.Run.Arch.X86_64]),
        ('64-BIT',     [gc3libs.Run.Arch.X86_64]),

        ('x86 32-bit', [gc3libs.Run.Arch.X86_32]),
        ('32bits x86', [gc3libs.Run.Arch.X86_32]),
        ('i386',       [gc3libs.Run.Arch.X86_32]),
        ('i486',       [gc3libs.Run.Arch.X86_32]),
        ('i586',       [gc3libs.Run.Arch.X86_32]),
        ('i686',       [gc3libs.Run.Arch.X86_32]),
        ('32 bits',    [gc3libs.Run.Arch.X86_32]),
        ('32-bit',     [gc3libs.Run.Arch.X86_32]),

        ('32bit, 64bit',  [gc3libs.Run.Arch.X86_64, gc3libs.Run.Arch.X86_32]),
        ('64bit, x86_64', [gc3libs.Run.Arch.X86_64]),
        ]
    config_template = gc3libs.template.Template("""
[resource/test]
architecture = ${arch}
""")
    for arch, result in test_cases:
        # display complete title for each generated test case
        _check_parse_arch.description = (
            test_valid_architectures.__doc__ + (': %s -> %s' % (str(arch), str(result))))
        yield (_check_parse_arch, config_template.substitute(arch=arch), result)

def _check_parse_arch(confstr, result):
    cfg = gc3libs.config.Configuration()
    defaults, resources, auths = cfg._parse(StringIO(confstr))
    assert_is_instance(resources['test']['architecture'], set)
    assert_equal(resources['test']['architecture'], set(result))


def test_invalid_architectures():
    """Test that invalid architecture strings are rejected with `ConfigurationError`"""
    test_cases = [
        '96bits', '31-BITS', 'amd65', 'xyzzy'
        ]
    config_template = gc3libs.template.Template("""
[resource/test]
architecture = ${arch}
""")
    for arch in test_cases:
        # display complete title for each generated test case
        _check_parse_arch.description = (
            test_invalid_architectures.__doc__ + (': %s' % arch))
        yield (raises(gc3libs.exceptions.ConfigurationError)(_check_parse_arch),
               config_template.substitute(arch=arch), 'should not be used')


def test_mandatory_configuration_options():
    """Check that mandatory configuration options are really required"""
    good_conf = """
[auth/ssh]
type = ssh
username = gc3pie

[resource/test]
# omit one line below
type = shellcmd
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
"""
    lines = good_conf.split('\n')
    header_ends_at = lines.index("# omit one line below")
    for omit in range(1+header_ends_at, len(lines)):
        if lines[omit].strip() == '':
            continue
        bad_conf = str.join('\n', lines[:omit] + lines[omit+1:])
        # display meaningful test name with `nosetests -v`
        _check_bad_conf.description = (
            test_mandatory_configuration_options.__doc__ + (" (omit '%s')" % lines[omit]))
        yield _check_bad_conf, bad_conf

@raises(TypeError, gc3libs.exceptions.ConfigurationError)
def _check_bad_conf(conf):
    tmp = _setup_config_file(conf)
    cfg = gc3libs.config.Configuration(tmp)
    # first check that errors are swallowed
    resources = cfg.make_resources()
    assert_equal(len(resources), 0)
    # next check that configuration error is raised
    try:
        resources = cfg.make_resources(ignore_errors=False)
    finally:
        os.remove(tmp)

class TestPrologueEpilogueScripts(object):
    """
    Test `prologue` and `epilogue` options for batch backends
    """
    def setUp(self):
        # setup conf dir and conf file
        self.files_to_remove = []
        cfgstring = """
[auth/ssh]
type = ssh
username = gc3pie

[resource/test]
type = shellcmd
auth = ssh
transport = local
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
prologue = scripts/shellcmd_pre.sh
epilogue = scripts/shellcmd_post.sh
myapp_prologue = scripts/myapp_shellcmd_pre.sh
myapp_epilogue = scripts/myapp_shellcmd_post.sh
override = False

[resource/testpbs]
type = pbs
auth = ssh
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
prologue = scripts/shellcmd_pre.sh
epilogue = scripts/shellcmd_post.sh
myapp_prologue = scripts/myapp_shellcmd_pre.sh
myapp_epilogue = scripts/myapp_shellcmd_post.sh
"""
        self.tmpdir = tempfile.mkdtemp()
        # setup config file and sctipts
        cfgfname = os.path.join(self.tmpdir, 'gc3pie.conf')
        self.files_to_remove.append(self.tmpdir)
        fdcfg = open(cfgfname, 'w')
        fdcfg.write(cfgstring)
        fdcfg.close()
        self.cfg = gc3libs.config.Configuration(cfgfname, auto_enable_auth=True)

        self.scripts = ['prologue', 'epilogue', 'myapp_prologue', 'myapp_epilogue']
        os.mkdir(os.path.join(self.tmpdir, 'scripts'))
        for k,v in self.cfg['resources']['test'].iteritems():
            if k in self.scripts:
                scriptfd = open(os.path.join(self.tmpdir, v), 'w')
                scriptfd.write('echo %s' % k)
                scriptfd.close()

        self.cfg = gc3libs.config.Configuration(cfgfname)
        # self.resources = self.cfg.make_resources()
        # assert_equal(resources['test']['prologue'], 'scripts/shellcmd_pre.sh')

    def tearDown(self):
        for dirname in self.files_to_remove:
            shutil.rmtree(dirname)

    def test_scriptfiles_are_abs(self):
        """Test that prologue and epilogue scripts are absolute pathnames"""
        self.core = gc3libs.core.Core(self.cfg)
        for resource in self.core.get_resources():
            for (k, v) in resource.iteritems():
                if 'prologue' not in k and 'epilogue' not in k:
                    continue
                assert os.path.isfile(v)
                assert os.path.isabs(v)
                assert_equal( os.path.abspath(v),
                              v)

    def test_pbs_prologue_and_epilogue_contents(self):
        """Test that prologue and epilogue scripts are correctly inserted into the submission script"""
        # Ugly hack. We have to list the job dirs to check which one
        # is the new one.
        jobdir = os.path.expanduser('~/.gc3pie_jobs')
        jobs = []
        if os.path.isdir(jobdir):
            jobs = os.listdir(jobdir)
        app = Application(['/bin/true'], [], [], '')
        self.core = gc3libs.core.Core(self.cfg)
        self.core.select_resource('testpbs')
        try:
            self.core.submit(app)
        except Exception, ex:
            # it is normal to have an error since we don't probably
            # run a pbs server in this machine.
            pass

        newjobs = [ d for d in os.listdir(jobdir) if d not in jobs]

        # There must be only one more job...
        assert_equal( len(newjobs), 1)

        newjobdir = os.path.join(jobdir, newjobs[0])
        self.files_to_remove.append(newjobdir)

        # and only one file in it
        assert_equal(len(os.listdir(newjobdir)), 1)

        # Check the content of the script file
        scriptfname = os.path.join(newjobdir, (os.listdir(newjobdir)[0]))
        scriptfile = open(scriptfname)
        assert re.match("#!/bin/sh.*# prologue file `.*` BEGIN.*echo prologue.*# prologue file END.*/bin/true.*# epilogue file `.*` BEGIN.*echo epilogue.*# epilogue file END", scriptfile.read(), re.DOTALL|re.M)
        scriptfile.close()

        # kill the job
        if app.execution.state != Run.State.NEW:
            self.core.kill(app)

    def test_prologue_epilogue_issue_352(self):
        cfgstring = """
[auth/ssh]
type = ssh
username = gc3pie

[resource/%s]
type = shellcmd
auth = ssh
transport = local
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
override = False
"""
        dir1, dir2 = tempfile.mkdtemp(), tempfile.mkdtemp()
        self.files_to_remove.extend([dir1, dir2])
        cfgfilenames = []

        for d in [dir1, dir2]:
            cfgfname = os.path.join(d, 'gc3pie.conf')
            fdcfg = open(cfgfname, 'w')
            fdcfg.write(cfgstring % os.path.basename(d))
            fdcfg.close()
            cfgfilenames.append(cfgfname)

        self.cfg = gc3libs.config.Configuration(
            cfgfilenames[0],
            self.cfg.cfgfiles[0],
            cfgfilenames[1], auto_enable_auth=True)

        self.core = gc3libs.core.Core(self.cfg)
        for resource in self.core.get_resources():
            for (k, v) in resource.iteritems():
                if 'prologue' not in k and 'epilogue' not in k:
                    continue
                assert os.path.isabs(v)
                assert os.path.isfile(v)
                assert_equal( os.path.abspath(v),
                              v)


if __name__ == "__main__":
    import nose
    nose.runmodule()



"""
Provide the `functools.wraps` function even if your Python doesn't
support it.  The rest of `functools` is only available if your Python
has it.
"""

try:
    from functools import *
except ImportError:
    pass

try:
    wraps
except NameError:
    def wraps(original):
        def inner(fn):
            # see functools.WRAPPER_ASSIGNMENTS
            for attribute in ['__module__',
                              '__name__',
                              '__doc__'
                              ]:
                setattr(fn, attribute, getattr(original, attribute))
            # see functools.WRAPPER_UPDATES
            for attribute in ['__dict__',
                              ]:
                if hasattr(fn, attribute):
                    getattr(fn, attribute).update(getattr(original, attribute))
                else:
                    setattr(fn, attribute,
                            getattr(original, attribute).copy())
            return fn
        return inner



"""
A backport of the Python standard `urlparse` package, providing
the `namedtuple` interface on Python 2.4.

See http://docs.python.org/release/2.7.2/library/urlparse.html
for a description of the module and a link to the original sources.
"""

__all__ = ["urlparse", "urlunparse", "urljoin", "urldefrag",
           "urlsplit", "urlunsplit", "parse_qs", "parse_qsl"]

import sys as _sys
if _sys.version >= (2, 5):
    # urlparse objects have attribute access since v2.5
    from urlparse import *

else:
    # use the source from Python 2.7 (SVN r87331):
    # http://svn.python.org/view/python/branches/release27-maint/Lib/urlparse.py?view=markup
        
    # A classification of schemes ('' means apply by default)
    uses_relative = ['ftp', 'http', 'gopher', 'nntp', 'imap',
                     'wais', 'file', 'https', 'shttp', 'mms',
                     'prospero', 'rtsp', 'rtspu', '', 'sftp']
    uses_netloc = ['ftp', 'http', 'gopher', 'nntp', 'telnet',
                   'imap', 'wais', 'file', 'mms', 'https', 'shttp',
                   'snews', 'prospero', 'rtsp', 'rtspu', 'rsync', '',
                   'svn', 'svn+ssh', 'sftp','nfs','git', 'git+ssh']
    non_hierarchical = ['gopher', 'hdl', 'mailto', 'news',
                        'telnet', 'wais', 'imap', 'snews', 'sip', 'sips']
    uses_params = ['ftp', 'hdl', 'prospero', 'http', 'imap',
                   'https', 'shttp', 'rtsp', 'rtspu', 'sip', 'sips',
                   'mms', '', 'sftp']
    uses_query = ['http', 'wais', 'imap', 'https', 'shttp', 'mms',
                  'gopher', 'rtsp', 'rtspu', 'sip', 'sips', '']
    uses_fragment = ['ftp', 'hdl', 'http', 'gopher', 'news',
                     'nntp', 'wais', 'https', 'shttp', 'snews',
                     'file', 'prospero', '']

    # Characters valid in scheme names
    scheme_chars = ('abcdefghijklmnopqrstuvwxyz'
                    'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                    '0123456789'
                    '+-.')

    MAX_CACHE_SIZE = 20
    _parse_cache = {}

    def clear_cache():
        """Clear the parse cache."""
        _parse_cache.clear()

    # use the `gc3libs.compat` namespace, so we can be sure this works
    # on Python 2.4 and 2.5
    from gc3libs.compat._collections import namedtuple
    
    class ResultMixin(object):
        """Shared methods for the parsed result objects."""

        @property
        def username(self):
            netloc = self.netloc
            if "@" in netloc:
                userinfo = netloc.rsplit("@", 1)[0]
                if ":" in userinfo:
                    userinfo = userinfo.split(":", 1)[0]
                return userinfo
            return None

        @property
        def password(self):
            netloc = self.netloc
            if "@" in netloc:
                userinfo = netloc.rsplit("@", 1)[0]
                if ":" in userinfo:
                    return userinfo.split(":", 1)[1]
            return None

        @property
        def hostname(self):
            netloc = self.netloc.split('@')[-1]
            if '[' in netloc and ']' in netloc:
                return netloc.split(']')[0][1:].lower()
            elif ':' in netloc:
                return netloc.split(':')[0].lower()
            elif netloc == '':
                return None
            else:
                return netloc.lower()

        @property
        def port(self):
            netloc = self.netloc.split('@')[-1].split(']')[-1]
            if ':' in netloc:
                port = netloc.split(':')[1]
                return int(port, 10)
            else:
                return None


    class SplitResult(namedtuple('SplitResult', 'scheme netloc path query fragment'), ResultMixin):
        
        __slots__ = ()

        def geturl(self):
            return urlunsplit(self)


    class ParseResult(namedtuple('ParseResult', 'scheme netloc path params query fragment'), ResultMixin):

        __slots__ = ()

        def geturl(self):
            return urlunparse(self)


    def urlparse(url, scheme='', allow_fragments=True):
        """Parse a URL into 6 components:
        <scheme>://<netloc>/<path>;<params>?<query>#<fragment>
        Return a 6-tuple: (scheme, netloc, path, params, query, fragment).
        Note that we don't break the components up in smaller bits
        (e.g. netloc is a single string) and we don't expand % escapes."""
        tuple = urlsplit(url, scheme, allow_fragments)
        scheme, netloc, url, query, fragment = tuple
        if scheme in uses_params and ';' in url:
            url, params = _splitparams(url)
        else:
            params = ''
        return ParseResult(scheme, netloc, url, params, query, fragment)

    def _splitparams(url):
        if '/'  in url:
            i = url.find(';', url.rfind('/'))
            if i < 0:
                return url, ''
        else:
            i = url.find(';')
        return url[:i], url[i+1:]

    def _splitnetloc(url, start=0):
        delim = len(url)   # position of end of domain part of url, default is end
        for c in '/?#':    # look for delimiters; the order is NOT important
            wdelim = url.find(c, start)        # find first of this delim
            if wdelim >= 0:                    # if found
                delim = min(delim, wdelim)     # use earliest delim position
        return url[start:delim], url[delim:]   # return (domain, rest)

    def urlsplit(url, scheme='', allow_fragments=True):
        """Parse a URL into 5 components:
        <scheme>://<netloc>/<path>?<query>#<fragment>
        Return a 5-tuple: (scheme, netloc, path, query, fragment).
        Note that we don't break the components up in smaller bits
        (e.g. netloc is a single string) and we don't expand % escapes."""
        allow_fragments = bool(allow_fragments)
        key = url, scheme, allow_fragments, type(url), type(scheme)
        cached = _parse_cache.get(key, None)
        if cached:
            return cached
        if len(_parse_cache) >= MAX_CACHE_SIZE: # avoid runaway growth
            clear_cache()
        netloc = query = fragment = ''
        i = url.find(':')
        if i > 0:
            if url[:i] == 'http': # optimize the common case
                scheme = url[:i].lower()
                url = url[i+1:]
                if url[:2] == '//':
                    netloc, url = _splitnetloc(url, 2)
                    if (('[' in netloc and ']' not in netloc) or
                            (']' in netloc and '[' not in netloc)):
                        raise ValueError("Invalid IPv6 URL")
                if allow_fragments and '#' in url:
                    url, fragment = url.split('#', 1)
                if '?' in url:
                    url, query = url.split('?', 1)
                v = SplitResult(scheme, netloc, url, query, fragment)
                _parse_cache[key] = v
                return v
            if url.endswith(':') or not url[i+1].isdigit():
                for c in url[:i]:
                    if c not in scheme_chars:
                        break
                else:
                    scheme, url = url[:i].lower(), url[i+1:]

        if url[:2] == '//':
            netloc, url = _splitnetloc(url, 2)
            if (('[' in netloc and ']' not in netloc) or
                    (']' in netloc and '[' not in netloc)):
                raise ValueError("Invalid IPv6 URL")
        if allow_fragments and scheme in uses_fragment and '#' in url:
            url, fragment = url.split('#', 1)
        if scheme in uses_query and '?' in url:
            url, query = url.split('?', 1)
        v = SplitResult(scheme, netloc, url, query, fragment)
        _parse_cache[key] = v
        return v

    def urlunparse(data):
        """Put a parsed URL back together again.  This may result in a
        slightly different, but equivalent URL, if the URL that was parsed
        originally had redundant delimiters, e.g. a ? with an empty query
        (the draft states that these are equivalent)."""
        scheme, netloc, url, params, query, fragment = data
        if params:
            url = "%s;%s" % (url, params)
        return urlunsplit((scheme, netloc, url, query, fragment))

    def urlunsplit(data):
        """Combine the elements of a tuple as returned by urlsplit() into a
        complete URL as a string. The data argument can be any five-item iterable.
        This may result in a slightly different, but equivalent URL, if the URL that
        was parsed originally had unnecessary delimiters (for example, a ? with an
        empty query; the RFC states that these are equivalent)."""
        scheme, netloc, url, query, fragment = data
        if netloc or (scheme and scheme in uses_netloc and url[:2] != '//'):
            if url and url[:1] != '/': url = '/' + url
            url = '//' + (netloc or '') + url
        if scheme:
            url = scheme + ':' + url
        if query:
            url = url + '?' + query
        if fragment:
            url = url + '#' + fragment
        return url

    def urljoin(base, url, allow_fragments=True):
        """Join a base URL and a possibly relative URL to form an absolute
        interpretation of the latter."""
        if not base:
            return url
        if not url:
            return base
        bscheme, bnetloc, bpath, bparams, bquery, bfragment = \
                urlparse(base, '', allow_fragments)
        scheme, netloc, path, params, query, fragment = \
                urlparse(url, bscheme, allow_fragments)
        if scheme != bscheme or scheme not in uses_relative:
            return url
        if scheme in uses_netloc:
            if netloc:
                return urlunparse((scheme, netloc, path,
                                   params, query, fragment))
            netloc = bnetloc
        if path[:1] == '/':
            return urlunparse((scheme, netloc, path,
                               params, query, fragment))
        if not path and not params:
            path = bpath
            params = bparams
            if not query:
                query = bquery
            return urlunparse((scheme, netloc, path,
                               params, query, fragment))
        segments = bpath.split('/')[:-1] + path.split('/')
        # XXX The stuff below is bogus in various ways...
        if segments[-1] == '.':
            segments[-1] = ''
        while '.' in segments:
            segments.remove('.')
        while 1:
            i = 1
            n = len(segments) - 1
            while i < n:
                if (segments[i] == '..'
                    and segments[i-1] not in ('', '..')):
                    del segments[i-1:i+1]
                    break
                i = i+1
            else:
                break
        if segments == ['', '..']:
            segments[-1] = ''
        elif len(segments) >= 2 and segments[-1] == '..':
            segments[-2:] = ['']
        return urlunparse((scheme, netloc, '/'.join(segments),
                           params, query, fragment))

    def urldefrag(url):
        """Removes any existing fragment from URL.

        Returns a tuple of the defragmented URL and the fragment.  If
        the URL contained no fragments, the second element is the
        empty string.
        """
        if '#' in url:
            s, n, p, a, q, frag = urlparse(url)
            defrag = urlunparse((s, n, p, a, q, ''))
            return defrag, frag
        else:
            return url, ''

    # unquote method for parse_qs and parse_qsl
    # Cannot use directly from urllib as it would create a circular reference
    # because urllib uses urlparse methods (urljoin).  If you update this function,
    # update it also in urllib.  This code duplication does not existin in Python3.

    _hexdig = '0123456789ABCDEFabcdef'
    _hextochr = dict((a+b, chr(int(a+b,16)))
                     for a in _hexdig for b in _hexdig)

    def unquote(s):
        """unquote('abc%20def') -> 'abc def'."""
        res = s.split('%')
        # fastpath
        if len(res) == 1:
            return s
        s = res[0]
        for item in res[1:]:
            try:
                s += _hextochr[item[:2]] + item[2:]
            except KeyError:
                s += '%' + item
            except UnicodeDecodeError:
                s += unichr(int(item[:2], 16)) + item[2:]
        return s

    def parse_qs(qs, keep_blank_values=0, strict_parsing=0):
        """Parse a query given as a string argument.

            Arguments:

            qs: percent-encoded query string to be parsed

            keep_blank_values: flag indicating whether blank values in
                percent-encoded queries should be treated as blank strings.
                A true value indicates that blanks should be retained as
                blank strings.  The default false value indicates that
                blank values are to be ignored and treated as if they were
                not included.

            strict_parsing: flag indicating what to do with parsing errors.
                If false (the default), errors are silently ignored.
                If true, errors raise a ValueError exception.
        """
        dict = {}
        for name, value in parse_qsl(qs, keep_blank_values, strict_parsing):
            if name in dict:
                dict[name].append(value)
            else:
                dict[name] = [value]
        return dict

    def parse_qsl(qs, keep_blank_values=0, strict_parsing=0):
        """Parse a query given as a string argument.

        Arguments:

        qs: percent-encoded query string to be parsed

        keep_blank_values: flag indicating whether blank values in
            percent-encoded queries should be treated as blank strings.  A
            true value indicates that blanks should be retained as blank
            strings.  The default false value indicates that blank values
            are to be ignored and treated as if they were  not included.

        strict_parsing: flag indicating what to do with parsing errors. If
            false (the default), errors are silently ignored. If true,
            errors raise a ValueError exception.

        Returns a list, as G-d intended.
        """
        pairs = [s2 for s1 in qs.split('&') for s2 in s1.split(';')]
        r = []
        for name_value in pairs:
            if not name_value and not strict_parsing:
                continue
            nv = name_value.split('=', 1)
            if len(nv) != 2:
                if strict_parsing:
                    raise ValueError, "bad query field: %r" % (name_value,)
                # Handle case of a control-name with no equal sign
                if keep_blank_values:
                    nv.append('')
                else:
                    continue
            if len(nv[1]) or keep_blank_values:
                name = unquote(nv[0].replace('+', ' '))
                value = unquote(nv[1].replace('+', ' '))
                r.append((name, value))

        return r
